"""
Need to configure a list of data sources from which to fetch files.  This is
done by setting the global *DATA_SOURCE* variable in the dataflow.fetch module

Each datasources needs to specify the name of the data source, its URL
and a path within the URL which will provide a list of files.  Alternatively,
if the data source has an associated DOI, then specify the DOI.

For example::

    from dataflow import fetch
    fetch.DATA_SOURCES = [
        {
            "name": "ncnr_DOI",
            "DOI": "10.18434/T4201B",
        },
        {
            "name": "ncnr",
            "url": "http://ncnr.nist.gov/pub/",
            "start_path": "ncnrdata",
        },
        {
            "name": "charlotte",
            "url": "http://charlotte.ncnr.nist.gov/pub",
            "start_path": "",
        },
        {
            "name": "nice_local",
            "url": "file:///",
            "start_path": "usr/local/nice/server_data/experiments",
        },
        {
            "name": "local",
            "url": "file:///",
            "start_path": "",
        },
    ]

"""

from __future__ import print_function

import datetime
import time
from posixpath import basename, join, sep
import hashlib

try:
    import urllib.request as urllib2
except ImportError:
    import urllib2

import pytz

from .calc import _format_ordered, generate_fingerprint
from .cache import get_file_cache
from .doi_resolve import get_target
from .lib.iso8601 import seconds_since_epoch

# override this if you want to point these to another place.
# in particular, remove the "local" option if deploying in the cloud!

DATA_SOURCES = []
DEFAULT_DATA_SOURCE = "ncnr"

def check_datasource(source):
    datasource = next((x for x in DATA_SOURCES if x['name'] == source), {})
    if datasource == {}:
        raise RuntimeError("Need to set dataflow.modules.load.DATA_SOURCES['" + source + "'] first!")
    if "url" in datasource:
        source_url = datasource["url"]
    elif "DOI" in datasource:
        source_url = get_target(datasource["DOI"])
        print("resolving DOI %s to url %s" % (datasource["DOI"], source_url))
        datasource["url"] = source_url # cache for next access
    else:
        source_url = ""
        raise RuntimeError("Must have url specified for data source: " + source + " in config.")
    return source_url


def url_get(fileinfo, mtime_check=True):
    path, mtime, entries = fileinfo['path'], fileinfo.get('mtime', None), fileinfo.get('entries', None)
    # fingerprint the get, leaving off entries information:
    cache = get_file_cache()
    fileinfo_minimal = {'path': path, 'mtime': mtime}
    config_str = str(_format_ordered(fileinfo_minimal))
    fp = generate_fingerprint(("url_get", config_str))
    if cache.exists(fp):
        ret = cache.get(fp)
        print("getting " + path + " from cache!")
    else:
        source = fileinfo.get("source", DEFAULT_DATA_SOURCE)
        name = basename(path)
        source_url = check_datasource(source)
        full_url = join(source_url, urllib2.quote(path.strip(sep), safe='/:'))
        url = None
        print("loading", full_url, name)
        try:
            url = urllib2.urlopen(full_url)
            if mtime_check:
                url_mtime = url.headers['last-modified']
                url_time_struct = time.strptime(url_mtime, '%a, %d %b %Y %H:%M:%S %Z')
                t_repo = datetime.datetime(*url_time_struct[:6], tzinfo=pytz.utc)
                t_request = datetime.datetime.fromtimestamp(mtime, pytz.utc)
                if mtime is None:
                    raise ValueError("timestamp checking enabled but no timestamp provided")
                elif t_request > t_repo:
                    print("request mtime = %s, repo mtime = %s"%(t_request, t_repo))
                    raise ValueError("Requested mtime is newer than repository mtime for %r"%path)
                elif t_request < t_repo:
                    print("request mtime = %s, repo mtime = %s"%(t_request, t_repo))
                    raise ValueError("Requested mtime is older than repository mtime for %r"%path)

            ret = url.read()
            print("caching " + path)
            cache.set(fp, ret)
        except urllib2.HTTPError as exc:
            raise ValueError("Could not open %r\n%s"%(path, str(exc)))
        finally:
            if url is not None:
                url.close()

    return ret

def find_mtime(path):
    check_datasource()
    try:
        url = urllib2.urlopen(DEFAULT_DATA_SOURCE+path)
        mtime = url.info().getdate('last-modified')
    except urllib2.HTTPError as exc:
        raise ValueError("Could not open %r\n%s"%(path, str(exc)))
    mtime_obj = datetime.datetime(*mtime[:7], tzinfo=pytz.utc)
    timestamp = seconds_since_epoch(mtime_obj)

    return {'path': path, 'mtime': timestamp}

def url_get_list(files=None):
    if files is None:
        return []
    result = [entry for fileinfo in files for entry in url_get(fileinfo)]
    return result



import os
import argparse

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--headless', action='store_true', help='do not automatically load client in browser')
    parser.add_argument('--external', action='store_true', help='listen on all interfaces, including external (local connections only if not set)')
    parser.add_argument('-p', '--port', default=8002, type=int, help='port on which to start the server')
    args = parser.parse_args()
    from reflweb.server_flask import app
    if not args.headless:
        import webbrowser
        webbrowser.open("http://localhost:%d" % (args.port))
    host = '0.0.0.0' if args.external else None
    app.run(port=args.port, host=host)

if __name__ == '__main__':
    main()



# This program is public domain

"""
Load a NeXus file into a reflectometry data structure.
"""
import os
import tempfile
from zipfile import ZipFile, _EndRecData
from io import BytesIO

import numpy as np
import h5py as h5

from dataflow.lib import hzf_readonly_stripped as hzf
from dataflow.lib import unit
from dataflow.lib import iso8601

from . import refldata
from .util import fetch_url

def data_as(group, fieldname, units, rep=1):
    """
    Return value of field in the desired units.
    """
    if fieldname not in group:
        return np.NaN
    field = group[fieldname]
    converter = unit.Converter(field.attrs.get('units', ''))
    value = converter(field.value, units)
    if rep != 1:
        if value.shape[0] == 1:
            return np.repeat(value, rep, axis=0)
        elif value.shape[0] != rep:
            raise ValueError("field %r does not match counts in %r"
                             %(field.name, field.file.filename))
        else:
            return value
    else:
        return value

def nxfind(group, nxclass):
    """
    Iterate over the entries of type *nxclass* in the hdf5 *group*.
    """
    for entry in group.values():
        if nxclass == entry.attrs.get('NX_class', None):
            yield entry


def load_from_string(filename, data, entries=None):
    """
    Load a nexus file from a string, e.g., as returned from url.read().
    """
    fd = BytesIO(data)
    entries = load_entries(filename, fd, entries=entries)
    fd.close()
    return entries


def load_from_uri(uri, entries=None, url_cache="/tmp"):
    """
    Load a nexus file from disk or from http://, https:// or file://.

    Remote files are cached in *url_cache*.  Use None to fetch without caching.
    """
    if uri.startswith('file://'):
        return load_entries(uri[7:], entries=entries)
    elif uri.startswith('http://') or uri.startswith('https://'):
        filename = os.path.basename(uri)
        data = fetch_url(uri, url_cache=url_cache)
        return load_from_string(filename, data, entries=entries)
    else:
        return load_entries(uri, entries=entries)


def load_metadata(filename, file_obj=None):
    """
    Load the summary info for all entries in a NeXus file.
    """
    #file = h5.File(filename)
    file = h5_open_zip(filename, file_obj)
    measurements = []
    for name, entry in file.items():
        if entry.attrs.get('NX_class', None) == 'NXentry':
            data = NCNRNeXusRefl(entry, name, filename)
            measurements.append(data)
    if file_obj is None:
        file.close()
    return measurements


def load_entries(filename, file_obj=None, entries=None):
    """
    Load the summary info for all entries in a NeXus file.
    """
    #file = h5.File(filename)
    file = h5_open_zip(filename, file_obj)
    measurements = []
    for name, entry in file.items():
        if entries is not None and name not in entries:
            continue
        if entry.attrs.get('NX_class', None) == 'NXentry':
            data = NCNRNeXusRefl(entry, name, filename)
            data.load(entry)
            measurements.append(data)
    if file_obj is None:
        file.close()
    return measurements

def h5_open_zip(filename, file_obj=None, mode='rb', **kw):
    """
    Open a NeXus file, even if it is in a zip file,
    or if it is a NeXus-zip file.

    If the filename ends in '.zip', it will be unzipped to a temporary
    directory before opening and deleted on :func:`closezip`.  If opened
    for writing, then the file will be created in a temporary directory,
    then zipped and deleted on :func:`closezip`.

    If it is a zipfile but doesn't end in '.zip', it is assumed
    to be a NeXus-zip file and is opened with that library.

    Arguments are the same as for :func:`open`.
    """
    if file_obj is None:
        file_obj = open(filename, mode=mode, buffering=-1)
    is_zip = _EndRecData(file_obj) # is_zipfile(file_obj) doens't work in py2.6
    if is_zip and '.attrs' in ZipFile(file_obj).namelist():
        # then it's a nexus-zip file, rather than
        # a zipped hdf5 nexus file
        f = hzf.File(filename, file_obj)
        f.delete_on_close = False
        f.zip_on_close = False
    else:
        zip_on_close = None
        if is_zip:
            path = tempfile.gettempdir()
            if mode == 'r':
                zf = ZipFile(filename)
                members = zf.namelist()
                assert len(members) == 1
                zf.extract(members[0], path)
                filename = os.path.join(path, members[0])
            elif mode == 'w':
                zip_on_close = filename
                filename = os.path.join(path, os.path.basename(filename)[:-4])
            else:
                raise TypeError("zipped nexus files only support mode r and w")

        f = h5.File(filename, mode=mode, **kw)
        f.delete_on_close = is_zip
        f.zip_on_close = zip_on_close
    return f

def h5_close_zip(f):
    """
    Close a NeXus file opened by :func:`open_zip`.

    If the file was zipped and opened for reading, delete the temporary
    file that was created
    If opened for writing, create a zip file containing the file
    before closing.

    Delete the file after closing.
    """
    path = f.filename
    delete_on_close = getattr(f, 'delete_on_close', False)
    zip_on_close = getattr(f, 'zip_on_close', None)
    f.close()
    if zip_on_close is not None:
        with ZipFile(f.zip_on_close, 'w') as zf:
            zf.write(path, os.path.basename(path))
    if delete_on_close:
        os.unlink(path)


class NCNRNeXusRefl(refldata.ReflData):
    """
    NeXus reflectometry entry.

    See :class:`refldata.ReflData` for details.
    """
    format = "NeXus"
    trajectory_intents = {
        'SPEC': 'specular',
        'SLIT': 'intensity',
        'BGP': 'background+',
        'BGM': 'background-',
        'ROCK': 'rock sample'
    }

    def __init__(self, entry, entryname, filename):
        super(NCNRNeXusRefl, self).__init__()
        self.entry = entryname
        self.path = os.path.abspath(filename)
        self._set_metadata(entry)

    def _set_metadata(self, entry):
        #print(entry['instrument'].values())
        das = entry['DAS_logs']
        self.probe = 'neutron'
        self.name = das['trajectoryData/fileName'][0] if 'trajectoryData/fileName' in das else 'unknown'
        if 'trajectoryData/fileNum' in das:
            self.filenumber = das['trajectoryData/fileNum'][0]
        else:
            # fall back to randomly generated filenum
            from random import randint
            self.filenumber = -randint(10**9, (10**10) - 1)

        #self.date = iso8601.parse_date(entry['start_time'][0].decode('utf-8'))
        self.date = iso8601.parse_date(entry['start_time'][0])
        self.description = entry['experiment_description'][0]
        self.instrument = entry['instrument/name'][0]
        self.slit1.distance = data_as(entry, 'instrument/presample_slit1/distance', 'mm')
        self.slit2.distance = data_as(entry, 'instrument/presample_slit2/distance', 'mm')
        #self.slit3.distance = data_as(entry, 'instrument/predetector_slit1/distance','mm')
        #self.slit4.distance = data_as(entry, 'instrument/predetector_slit2/distance','mm')
        #self.detector.distance = data_as(entry, 'instrument/detector/distance','mm')
        #self.detector.rotation = data_as(entry, 'instrument/detector/rotation','degree')
        self.detector.deadtime = data_as(entry, 'instrument/single_detector/dead_time', 'us')
        self.detector.deadtime_error = data_as(entry, 'instrument/single_detector/dead_time_error', 'us')
        monitor_device = entry.get('control/monitor', {})
        self.monitor.deadtime = data_as(monitor_device, 'dead_time','us')
        self.monitor.deadtime_error = data_as(monitor_device, 'dead_time_error', 'us')

        self.sample.name = entry['sample/name'][0] if 'name' in entry['sample'] else ""
        self.sample.description = entry['sample/description'][0] if 'description' in entry['sample'] else ""
        raw_intent = das['trajectoryData/_scanType'][0] if 'trajectoryData/_scanType' in das else ""
        if raw_intent in self.trajectory_intents:
            self.intent = self.trajectory_intents[raw_intent]
        self.monitor.base = das['counter/countAgainst'][0]
        self.monitor.time_step = 0.001  # assume 1 ms accuracy on reported clock
        self.polarization = _get_pol(das, 'frontPolarization') \
                            + _get_pol(das, 'backPolarization')

        if np.isnan(self.slit1.distance):
            self.warn("Slit 1 distance is missing; using 2 m")
            self.slit1.distance = -2000
        if np.isnan(self.slit2.distance):
            self.warn("Slit 2 distance is missing; using 1 m")
            self.slit2.distance = -1000
        if np.isnan(self.detector.wavelength):
            self.warn("Wavelength is missing; using 4.75 A")
            self.detector.wavelength = 4.75
        if np.isnan(self.detector.wavelength_resolution):
            self.warn("Wavelength resolution is missing; using 1.5% dL/L FWHM")
            self.detector.wavelength_resolution = 0.015/2.35*self.detector.wavelength

        # TODO: stop trying to guess DOI
        if 'DOI' in entry:
            URI = entry['DOI']
        else:
            # See: dataflow.modules.doi_resolve for helpers.
            #NCNR_DOI = "10.18434/T4201B"
            NCNR_DOI = "https://ncnr.nist.gov/pub/ncnrdata"
            LOCATION = {'pbr':'ngd', 'magik':'cgd', 'ng7r':'ng7'}
            nice_instrument = str(das['experiment/instrument'].value[0]).lower()
            instrument = LOCATION.get(nice_instrument, nice_instrument)
            year, month = self.date.year, self.date.month
            cycle = "%4d%02d"%(year, month)
            experiment = str(entry['experiment_identifier'].value[0])
            filename = os.path.basename(self.path)
            URI = "/".join((NCNR_DOI, instrument, cycle, experiment, "data", filename))
        self.uri = URI

    def load(self, entry):
        das = entry['DAS_logs']
        self.detector.counts = np.asarray(data_as(das, 'counter/liveROI', ''), 'd')
        self.detector.counts_variance = self.detector.counts.copy()
        self.detector.dims = self.detector.counts.shape
        n = self.detector.dims[0]
        self.points = n
        self.monitor.counts = np.asarray(data_as(das, 'counter/liveMonitor', '', rep=n), 'd')
        self.monitor.counts_variance = self.monitor.counts.copy()
        self.monitor.count_time = data_as(das, 'counter/liveTime', 's', rep=n)
        self.detector.wavelength = data_as(entry, 'instrument/monochromator/wavelength','Ang', rep=n)
        self.detector.wavelength_resolution = data_as(entry, 'instrument/monochromator/wavelength_error','Ang', rep=n)
        for k, s in enumerate([self.slit1, self.slit2, self.slit3, self.slit4]):
            x = 'slitAperture%d/softPosition'%(k+1)
            y = 'vertSlitAperture%d/softPosition'%(k+1)
            x_target = 'slitAperture%d/desiredSoftPosition'%(k+1)
            y_target = 'vertSlitAperture%d/desiredSoftPosition'%(k+1)
            if x in das:
                s.x = data_as(das, x, 'mm', rep=n)
            if y in das:
                s.y = data_as(das, y, 'mm', rep=n)
            s.x_target = data_as(das, x_target, 'mm', rep=n)
            s.y_target = data_as(das, y_target, 'mm', rep=n)
        if 'sampleAngle' in das:
            # selects MAGIK or PBR, which have sample and detector angle
            self.sample.angle_x = data_as(das, 'sampleAngle/softPosition', 'degree', rep=n)
            self.detector.angle_x = data_as(das, 'detectorAngle/softPosition', 'degree', rep=n)
            self.sample.angle_x_target = data_as(das, 'sampleAngle/desiredSoftPosition', 'degree', rep=n)
            self.detector.angle_x_target = data_as(das, 'detectorAngle/desiredSoftPosition', 'degree', rep=n)
        elif 'q' in das:
            # selects NG7R which has only q device (qz) and sampleTilt
            # Ignore sampleTilt for now since it is arbitrary.  NG7 is not
            # using zeros for the sampleTilt motor in a predictable way.
            tilt = 0.
            #tilt = data_as(das, 'sampleTilt/softPosition', 'degree', rep=n)
            theta = data_as(das, 'q/thetaIncident', 'degree', rep=n)
            self.sample.angle_x = theta + tilt
            self.detector.angle_x = 2*theta
            # NaN if not defined
            tilt_target = 0.
            #tilt_target = data_as(das, 'sampleTilt/desiredSoftPosition', 'degree', rep=n)
            theta_target = data_as(das, 'q/desiredThetaInident', 'degree', rep=n)
            self.sample.angle_x_target = theta_target + tilt_target
            self.detector.angle_x_target = 2*theta_target
        else:
            raise ValueError("Unknown sample angle in file")
        self.Qz_target = data_as(das, 'trajectoryData/_q', '', rep=n)
        if 'trajectoryData/_theta_offset' in das:
            self.background_offset = 'theta'
        self.scan_value = []
        self.scan_units = []
        self.scan_label = []
        SCANNED_VARIABLES = 'trajectory/scannedVariables'
        if SCANNED_VARIABLES in das:
            scanned_variables = das[SCANNED_VARIABLES].value
            # Just in case the scanned variables is a string with
            # elements separated by new lines...
            if len(scanned_variables) == 1:
                scanned_variables = str(scanned_variables[0]).split('\n')
            scanned_variables = [s for s in scanned_variables if not s.startswith("areaDetector")]
            for node_id in scanned_variables:
                path = node_id.replace('.', '/')
                try:
                    field = das[path]
                except KeyError:
                    print(">>> could not read scanned %s for %s"
                          % (node_id, os.path.basename(self.path)))
                    continue
                try:
                    scan_value = data_as(das, path, '', rep=n)
                    scan_units = field.attrs.get('units', '')
                    scan_label = field.attrs.get('label', node_id)
                except Exception:
                    print(">>> unexpected error reading %s for %s"
                          % (node_id, os.path.basename(self.path)))
                    continue
                # check if numeric:
                if scan_value.dtype.kind in ["f", "u", "i"]:
                    self.scan_value.append(scan_value)
                    self.scan_units.append(scan_units)
                    self.scan_label.append(scan_label)
        # TODO: field
        if 'temp' in das:
            if 'temp/primaryControlLoop' in das:
                temp_primaryControlLoop = das['temp/primaryControlLoop'].value
                self.sample.temp_setpoint = data_as(das, 'temp/setpoint_%d' % (temp_primaryControlLoop), 'K')[0]
            temp_values = data_as(das, 'temp/primaryNode/value', 'K')
            temp_shape = temp_values.shape[0] if temp_values.shape[0] else 1.0;
            # only include one significant figure for temperatures.
            self.sample.temp_avg = round(np.sum(temp_values)/temp_shape, 1)

    def _load_slits(self, instrument):
        """
        Slit names have not been standardized.  Instead sort the
        NXaperature components by distance and assign them according
        to serial order, negative aperatures first and positive second.
        """
        raise NotImplementedError("hardcode slit names for now")
        slits = list(nxfind(instrument, 'NXaperature'))
        # Note: only supports to aperatures before and after.
        # Assumes x and y aperatures are coupled in the same
        # component.  This will likely be wrong for some instruments,
        # but we won't deal with that until we have real NeXus files
        # to support.
        # Assume the file writer was sane and specified all slit
        # distances in the same units so that sorting is simple.
        # Currently only slit distance is recorded, not slit opening.
        slits.sort(lambda a,b: -1 if a.distance < b.distance else 1)
        index = 0
        for slit in slits:
            d = slit.distance.value('meters')
            if d <= 0:
                # process first two slits only
                if index == 0:
                    self.slit1.distance = d
                    index += 1
                elif index == 1:
                    self.slit2.distance = d
                    index += 1
            elif d > 0:
                # skip leading slits
                if index < 2: index = 2
                if index == 2:
                    self.slit3.distance = d
                    index += 1
                elif index == 3:
                    self.slit4.distance = d
                    index += 1

def _get_pol(das, pol):
    if pol in das:
        direction = das[pol+'/direction'][0]
        if direction == 'UP':
            result = '+'
        elif direction == 'DOWN':
            result = '-'
        elif direction == '' or direction == 'BEAM_OUT' or direction == 'UNPOLARIZED':
            result = ''
        else:
            raise ValueError("Don't understand DAS_logs/%s/direction=%r"%(pol,direction))
    else:
        result = ''
    return result



def demo():
    import sys
    if len(sys.argv) == 1:
        print("usage: python -m reflred.steps.nexusref file...")
        sys.exit(1)
    for filename in sys.argv[1:]:
        try:
            entries = load_from_uri(filename)
        except Exception as exc:
            print("**** "+str(exc)+" **** while reading "+filename)
            continue

        # print the first entry
        print(entries[0])

        # plot all the entries
        import pylab
        pylab.figure()
        for entry in entries:
            entry.plot()
        pylab.legend()
        pylab.show()

if __name__ == "__main__":
    demo()



"""
1-D and 2-D rebinning code.
"""
import numpy as np

try:
    from . import _reduction  ## C API wrapper
    #from . import _rebin  ## cython wrapper
except ImportError:
    import warnings
    warnings.warn("Reduction module not compiled...rebinning unavailable")

try:
    #from typing import Sequence, Optional, Union
    pass
except ImportError:
    pass

def rebin(x, I, xo, Io=None, dtype=np.float64):
    # type: (Sequence, Sequence, Sequence, Optional[np.ndarray], Union[str, type]) -> np.ndarray
    """
    Rebin a vector.

    x are the existing bin edges
    xo are the new bin edges
    I are the existing counts (one fewer than edges)

    Io will be used if present, but be sure that it is a contiguous
    array of the correct shape and size.

    dtype is the type to use for the intensity vectors.  This can be
    integer (uint8, uint16, uint32) or real (float32 or f, float64 or d).
    The edge vectors are all coerced to doubles.

    Note that total intensity is not preserved for integer rebinning.
    The algorithm uses truncation so total intensity will be down on
    average by half the total number of bins.
    """
    from .rebin_python import rebin_counts
    # Coerce axes to float arrays
    x, xo = _input(x, dtype='d'), _input(xo, dtype='d')
    shape_in = np.array([x.shape[0]-1])
    shape_out = np.array([xo.shape[0]-1])

    # Coerce counts to correct type and check shape
    if dtype is None:
        dtype = getattr(I, 'dtype', np.float64)
    I = _input(I, dtype=dtype)
    if shape_in != I.shape:
        raise TypeError("input array incorrect shape %s"%I.shape)

    # Create output vector
    Io = _output(Io, shape_out, dtype=dtype)

    rebin_counts(x, I, xo, Io)
    return Io

def rebin2d(x, y, I, xo, yo, Io=None, dtype=None):
    # type: (Sequence, Sequence, Sequence, Sequence, Sequence, Optional[np.ndarray], Union[str, type]) -> np.ndarray
    """
    Rebin a matrix.

    x, y are the existing bin edges
    xo, yo are the new bin edges
    I is the existing counts (one fewer than edges in each direction)

    For example, with x representing the column edges in each row and
    y representing the row edges in each column, the following
    represents a uniform field::

        >>> from reflred.rebin import rebin2d
        >>> x, y = [0, 2, 4, 5], [0, 1, 3]
        >>> z = [[2, 2, 1], [4, 4, 2]]

    We can check this by rebinning with uniform size bins::

        >>> xo, yo = range(6), range(4)
        >>> rebin2d(y, x, z, yo, xo)
        array([[ 1.,  1.,  1.,  1.,  1.],
               [ 1.,  1.,  1.,  1.,  1.],
               [ 1.,  1.,  1.,  1.,  1.]])

    dtype is the type to use for the intensity vectors.  This can be
    integer (uint8, uint16, uint32) or real (float32 or f, float64 or d).
    The edge vectors are all coerced to doubles.

    Note that total intensity is not preserved for integer rebinning.
    The algorithm uses truncation so total intensity will be down on
    average by half the total number of bins.

    Io will be used if present, if it is contiguous and if it has the
    correct shape and type for the input.  Otherwise it will raise a
    TypeError.  This will allow you to rebin the slices of an appropriately
    ordered matrix without making copies.
    """
    from .rebin_python import rebin_counts_2D
    # Coerce axes to float arrays
    x, y, xo, yo = [_input(v, dtype='d') for v in (x, y, xo, yo)]
    shape_in = np.array([x.shape[0]-1, y.shape[0]-1])
    shape_out = np.array([xo.shape[0]-1, yo.shape[0]-1])

    # Coerce counts to correct type and check shape
    if dtype is None:
        dtype = getattr(I, 'dtype', np.float64)
    I = _input(I, dtype=dtype)
    if (shape_in != I.shape).any():
        raise TypeError("input array incorrect shape %s"%str(I.shape))

    # Create output vector
    Io = _output(Io, shape_out, dtype=dtype)

    rebin_counts_2D(x, y, I, xo, yo, Io)
    return Io

def _input(v, dtype='d'):
    # type: Sequence -> np.ndarray
    """
    Force v to be a contiguous array of the correct type, avoiding copies
    if possible.
    """
    return np.ascontiguousarray(v, dtype=dtype)

def _output(v, shape, dtype=np.float64):
    """
    Create a contiguous array of the correct shape and type to hold a
    returned array, reusing an existing array if possible.
    """
    if v is None:
        return np.empty(shape, dtype=dtype)
    if not (isinstance(v, np.ndarray)
            and v.dtype == np.dtype(dtype)
            and v.shape == shape
            and v.flags.contiguous):
        raise TypeError("output vector must be contiguous %s of size %s"
                        %(dtype, shape))
    return v


# ================ Test code ==================
# TODO: move test code to its own file
def _check1d(from_bins, val, to_bins, target):
    """
    Test 1D rebinning *from_bins* => *to_bins* of counts *val* => *target*.

    The expected *target* must be precomputed in order to test that the right
    value is produced.

    Tests all combinations of increasing/decreasing source/target bin edges.

    Yields a sequence of test functions *t() -> None* which raise errors if
    the result is unexpected.  *t.description* is set to the name of the
    test variant so that nosetests will have slightly nicer error reporting.
    """
    target = _input(target)
    for (f, F) in [(from_bins, val), (from_bins[::-1], val[::-1])]:
        for (t, T) in [(to_bins, target), (to_bins[::-1], target[::-1])]:
            test_name = "rebin %s->%s: %s->%s"%(f, t, F, T)
            def test():
                result = rebin(f, F, t)
                assert np.linalg.norm(T-result) < 1e-14, \
                    "rebin failed %s: %s "%(test_name, result)
            test.description = test_name
            yield test

def _test1d():
    # Split a value
    for t in _check1d([1, 2, 3, 4], [10, 20, 30], [1, 2.5, 4], [20, 40]):
        yield t

    # bin is a superset of rebin
    for t in _check1d([0, 1, 2, 3, 4], [5, 10, 20, 30], [1, 2.5, 3], [20, 10]):
        yield t

    # bin is a subset of rebin
    for t in _check1d([1, 2, 3, 4, 5, 6], [10, 20, 30, 40, 50], [2.5, 3.5], [25]):
        yield t

    # one bin to many
    for t in _check1d([1, 2, 3, 4, 5, 6], [10, 20, 30, 40, 50], [2.1, 2.2, 2.3, 2.4], [2, 2, 2]):
        yield t

    # many bins to one
    for t in _check1d([1, 2, 3, 4, 5, 6], [10, 20, 30, 40, 50], [2.5, 4.5], [60]):
        yield t

def _check2d(x, y, z, xo, yo, zo):
    # type: (Sequence, Sequence, Sequence, Sequence, Sequence, Sequence) -> None
    result = rebin2d(x, y, z, xo, yo)
    target = np.array(zo, dtype=result.dtype)
    assert np.linalg.norm(target-result) < 1e-14, \
        "rebin2d failed for %s, %s->%s, %s\n%s\n%s" % (x, y, xo, yo, z, zo)

def _uniform_test(x, y):
    z = np.array([y], 'd') * np.array([x], 'd').T
    xedges = np.concatenate([(0, ), np.cumsum(x)])
    yedges = np.concatenate([(0, ), np.cumsum(y)])
    nx = np.round(xedges[-1])
    ny = np.round(yedges[-1])
    ox = np.arange(nx+1)
    oy = np.arange(ny+1)
    target = np.ones([nx, ny], 'd')
    return _check2d(xedges, yedges, z, ox, oy, target)

def _test2d():
    x, y, I = [0, 3, 5, 7], [0, 1, 3], [[3, 6], [2, 4], [2, 4]]
    xo, yo, Io = range(8), range(4), [[1]*3]*7
    x, y, I, xo, yo, Io = [np.array(A, 'd') for A in [x, y, I, xo, yo, Io]]

    # Try various types and orders on a non-square matrix
    yield lambda: _check2d(x, y, I, xo, yo, Io)
    yield lambda: _check2d(x[::-1], y, I[::-1, :], xo, yo, Io)
    yield lambda: _check2d(x, y[::-1], I[:, ::-1], xo, yo, Io)
    yield lambda: _check2d(x, y, I, [7, 3, 0], yo, [[4]*3, [3]*3])
    yield lambda: _check2d(x, y, I, xo, [3, 2, 0], [[1, 2]]*7)
    yield lambda: _check2d(y, x, I.T, yo, xo, Io.T)  # C vs. Fortran ordering

    # Test smallest possible result
    yield lambda: _check2d([-1, 2, 4], [0, 1, 3],
                           [[3, 6], [2, 4]], [1, 2], [1, 2], [1])
    # subset/superset
    yield lambda: _check2d([0, 1, 2, 3], [0, 1, 2, 3],
                           [[1]*3]*3, [0.5, 1.5, 2.5], [0.5, 1.5, 2.5], [[1]*2]*2)
    #for dtype in ['uint8', 'uint16', 'uint32', 'uint64', 'float32', 'float64']:
    #    yield lambda: _check2d(
    #        [0, 1, 2, 3, 4], [0, 1, 2, 3, 4],
    #        np.array([[1]*4]*4, dtype=dtype),
    #        [-2, -1, 2, 5, 6], [-2, -1, 2, 5, 6],
    #        np.array([[0, 0, 0, 0], [0, 4, 4, 0], [0, 4, 4, 0], [0, 0, 0, 0]], dtype=dtype)
    #        )
    # non-square test
    #yield lambda: _uniform_test([1, 2.5, 4, 0.5], [3, 1, 2.5, 1, 3.5])
    #yield lambda: _uniform_test([3, 2], [1, 2])

def test():
    for t in _test1d(): yield t
    for t in _test2d(): yield t

def main():
    for t in test(): t()

if __name__ == "__main__":
    main()



import numpy as np

from dataflow.lib import err1d


def apply_rescale(data, scale, dscale):
    I, varI = err1d.mul(data.v, data.dv**2, scale, dscale**2)
    data.v, data.dv = I, np.sqrt(varI)


def apply_intensity_norm(data, base):
    assert data.normbase == base.normbase, "can't mix time and monitor normalized data"
    S, varS = err1d.interp(data.angular_resolution,
                           base.angular_resolution, base.v, base.dv**2)
    I, varI = err1d.div(data.v, data.dv**2, S, varS)
    data.v, data.dv = I, np.sqrt(varI)
    
def calculate_number(data, base, time_uncertainty=1e-6):
    """ returns the measured base flux * count time for each point """
    assert base.normbase == 'time', "can't calculate time-integrated flux from monitor-normalized base"
    S, varS = err1d.interp(data.angular_resolution,
                           base.angular_resolution, base.v, base.dv**2)
    F, varF = err1d.mul(data.monitor.count_time, time_uncertainty**2, S, varS)
    return F, varF

def estimate_attenuation(datasets):
    raise NotImplementedError()
    index = np.sort([d.angular_resolution[0] for d in datasets])


NORMALIZE_OPTIONS = 'auto|monitor|time|power|none'
def apply_norm(data, base='auto'):
    if base == 'auto':
        if (data.monitor.counts > 0).all():
            base = 'monitor'
        elif (data.monitor.count_time > 0).all():
            base = 'time'
        elif data.monitor.source_power is not None:
            base = 'power'
        else:
            base = 'none'

    C = data.detector.counts
    varC = data.detector.counts_variance
    if base == 'monitor':
        assert (data.monitor.counts > 0).all(), "monitor counts are zero; can't normalize by monitor"
        M = data.monitor.counts
        varM = data.monitor.counts_variance
        varM += (varM == 0)  # variance on zero counts is +/- 1
        units = 'monitor'
    elif base == 'time':
        assert (data.monitor.count_time > 0).all(), "count time is zero; can't normalize by time"
        M = data.monitor.count_time
        # Uniform distribution has variance of interval/12
        varM = data.monitor.time_step/12.
        units = 'second'
    elif base == 'power':
        assert data.monitor.source_power is not None, "source power is unknown; can't normalize by power"
        M = data.monitor.source_power
        varM = 0
        units = data.monitor.source_power_units
    elif base == 'none':
        M = 1
        varM = 0
        units = ''
    else:
        raise ValueError("Expected %r in %s" % (base, NORMALIZE_OPTIONS))
    #print "norm",C,varC,M,varM
    value, variance = err1d.div(C, varC+(varC == 0), M, varM)
    data.v = value
    data.dv = np.sqrt(variance)
    data.vunits = 'counts per '+units if units else 'counts'
    data.vlabel = 'Intensity'
    data.normbase = base

    # TODO: set data scale to a nice number
    #data.scale = nice_number(M[0])




"""
Based on scalars or numpy vectors, this class allows you to store and
manipulate values+uncertainties, with propagation of gaussian error for
addition, subtraction, multiplication, division, power, exp, log
and trig.  Also includes mean, weighted average, and linear interpolation.

Storage properties are determined by the numbers used to set the value
and uncertainty.  Be sure to use floating point uncertainty vectors
for inplace operations since numpy does not do automatic type conversion.
Normal operations can use mixed integer and floating point.  In place
operations (a*=b, etc.) create at most one extra copy for each operation.
The copy operation c=a*b by contrast uses four intermediate vectors, so
shouldn't be used for huge arrays.
"""
from __future__ import division

__all__ = ['Uncertainty']

from copy import copy

import numpy as np

from . import err1d
from .formatnum import format_uncertainty
from . import wsolve


# TODO: rename to Measurement and add support for units?
# TODO: C implementation of *, /, **?
class Uncertainty(object):
    __slots__ = ('x', 'variance')
    __array_priority__ = 20.   # force array*uncertainty to use our __rmul__

    # Constructor
    def __init__(self, x, variance=None):
        self.x, self.variance = x, variance

    def __copy__(self):
        return Uncertainty(copy(self.x), copy(self.variance))

    @property
    def dtype(self):
        return self.x.dtype
    @property
    def flat(self):
        return iter(self)
    @property
    def ndim(self):
        return self.x.ndim
    @property
    def shape(self):
        return self.x.shape
    @property
    def size(self):
        return self.x.size
    @property
    def T(self):
        return self.transpose()
    # undefined properties:
    #    base, ctypes, data, flags, itemsize, nbytes, strides

    # Comparisons are not well defined for uncertainty objects.  For example,
    # is it true that (3 +/- 0.1) > (2 +/- 10)?  Running a quick simulation,
    # it is only true 54% of the time, so it is neither True nor False.
    # Similarly for equality/inequality.  The following functions are therefore
    # ill-defined:
    #     >, >=, <, <=, ==, !=, bool, all, any, nonzero
    #     argmax, amax, nanmax, maximum, fmax
    #     argmin, amin, nanmin, minimum, fmin
    #     argsort, sort, argpartition, partition, searchsorted
    #     clip, ptp
    # Could perhaps set a p-value for all operations, defaulting to p=0.05,
    # and using two-tailed tests.  That would at least give a partial ordering.
    # Bool, all, and any would involve a comparison against zero.  Maximum
    # likelihood could be used for maximum/minimum, and for the comparisons
    # used for sorting, which amounts to simply using the x value and ignoring
    # the variance.  For now, leave it undefined.

    # getfield, setfield, astype, byteswap, newbyteorder, setflags: use A.x.op

    # TODO: dump, dumps
    # TODO: item, itemset
    # TODO: tobytes, tofile, tolist, tostring

    # choose: index array will not be an uncertainty object, so no choose
    # method, but still may want to implement a choose function.  What happens
    # with np.choose when choices are all uncertainty objects?

    # cumsum mean, sum, std, var: see below
    # TODO: cumprod, prod, dot, trace

    # TODO: round() could truncate x to the number of digits defined by the
    # standard deviation and truncate variance to two digits.

    # TODO: assuming real, imag are identically distributed.  Is that true?
    def conj(self):
        return Uncertainty(self.x.conj(), self.variance)
    conjugate = conj
    def real(self):
        return Uncertainty(self.x.real(), self.variance)
    def imag(self):
        return Uncertainty(self.x.imag(), self.variance)

    def fill(self, a):
        if isinstance(a, Uncertainty):
            self.x.fill(a.x)
            self.variance.fill(a.variance)
        else:
            self.x.fill(a)
            self.variance.fill(0.)

    # Array shaping/selection operations
    def compress(self, *args, **kw):
        # TODO: if out is supplied, then split it into out.x, out.variance
        return Uncertainty(self.x.compress(*args, **kw),
                           self.variance.compress(*args, **kw))
    copy = __copy__
    def diagonal(self, *args, **kw):
        return Uncertainty(self.x.diagonal(*args, **kw),
                           self.variance.diagonal(*args, **kw))
    def ravel(self, *args, **kw):
        return Uncertainty(self.x.ravel(*args, **kw),
                           self.variance.ravel(*args, **kw))
    def reshape(self, *args, **kw):
        return Uncertainty(self.x.reshape(*args, **kw),
                           self.variance.reshape(*args, **kw))
    def repeat(self, *args, **kw):
        return Uncertainty(self.x.repeat(*args, **kw),
                           self.variance.repeat(*args, **kw))
    def resize(self, *args, **kw):
        return Uncertainty(self.x.resize(*args, **kw),
                           self.variance.resize(*args, **kw))
    def squeeze(self, *args, **kw):
        return Uncertainty(self.x.squeeze(*args, **kw),
                           self.variance.squeeze(*args, **kw))
    def swapaxes(self, *args, **kw):
        return Uncertainty(self.x.swapaxes(*args, **kw),
                           self.variance.swapaxes(*args, **kw))
    def take(self, *args, **kw):
        return Uncertainty(self.x.take(*args, **kw),
                           self.variance.take(*args, **kw))
    def transpose(self, *args, **kw):
        return Uncertainty(self.x.transpose(*args, **kw),
                           self.variance.transpose(*args, **kw))
    def view(self, *args, **kw):
        return Uncertainty(self.x.view(*args, **kw),
                           self.variance.view(*args, **kw))

    # Make standard deviation available
    def _getdx(self):
        return np.sqrt(self.variance)
    def _setdx(self, dx):
        # Direct operation
        #    variance = dx**2
        # Indirect operation to avoid temporaries
        self.variance[:] = dx
        self.variance **= 2
    dx = property(_getdx, _setdx, doc="standard deviation")

    def rand(self, size=None):
        return np.random.randn(size=size)*self.dx + self.x

    # Numpy array slicing operations
    def __len__(self):
        return len(self.x)
    def __getitem__(self, key):
        return Uncertainty(self.x[key], self.variance[key])
    def __setitem__(self, key, value):
        self.x[key] = value.x
        self.variance[key] = value.variance
    def __delitem__(self, key):
        del self.x[key]
        del self.variance[key]
    def __iter__(self):
        for x, variance in zip(self.x.flat, self.variance.flat):
            yield Uncertainty(x, variance)

    # Normal operations: may be of mixed type
    def __add__(self, other):
        if isinstance(other, Uncertainty):
            return Uncertainty(*err1d.add(self.x, self.variance, other.x, other.variance))
        else:
            return Uncertainty(self.x+other, self.variance+0) # Force copy
    def __sub__(self, other):
        if isinstance(other, Uncertainty):
            return Uncertainty(*err1d.sub(self.x, self.variance, other.x, other.variance))
        else:
            return Uncertainty(self.x-other, self.variance+0) # Force copy
    def __mul__(self, other):
        if isinstance(other, Uncertainty):
            return Uncertainty(*err1d.mul(self.x, self.variance, other.x, other.variance))
        else:
            return Uncertainty(self.x*other, self.variance*other**2)
    def __truediv__(self, other):
        if isinstance(other, Uncertainty):
            return Uncertainty(*err1d.div(self.x, self.variance, other.x, other.variance))
        else:
            return Uncertainty(self.x/other, self.variance/other**2)
    def __pow__(self, other):
        if isinstance(other, Uncertainty):
            return Uncertainty(*err1d.pow2(self.x, self.variance, other.x, other.variance))
        else:
            return Uncertainty(*err1d.pow(self.x, self.variance, other))

    # Reverse operations are definitely of mixed type
    def __radd__(self, other):
        return Uncertainty(self.x+other, self.variance+0) # Force copy
    def __rsub__(self, other):
        return Uncertainty(other-self.x, self.variance+0)
    def __rmul__(self, other):
        return Uncertainty(self.x*other, self.variance*other**2)
    def __rtruediv__(self, other):
        return Uncertainty(other/self.x, self.variance*other**2/self.x**4)
    def __rpow__(self, other):
        return exp(np.log(other)*self)

    # In-place operations: may be of mixed type
    def __iadd__(self, other):
        if isinstance(other, Uncertainty):
            self.x, self.variance \
                = err1d.add_inplace(self.x, self.variance, other.x, other.variance)
        else:
            self.x += other
        return self
    def __isub__(self, other):
        if isinstance(other, Uncertainty):
            self.x, self.variance \
                = err1d.sub_inplace(self.x, self.variance, other.x, other.variance)
        else:
            self.x -= other
        return self
    def __imul__(self, other):
        if isinstance(other, Uncertainty):
            self.x, self.variance \
                = err1d.mul_inplace(self.x, self.variance, other.x, other.variance)
        else:
            self.x *= other
            self.variance *= other**2
        return self
    def __itruediv__(self, other):
        if isinstance(other, Uncertainty):
            self.x, self.variance \
                = err1d.div_inplace(self.x, self.variance, other.x, other.variance)
        else:
            self.x /= other
            self.variance /= other**2
        return self
    def __ipow__(self, other):
        if isinstance(other, Uncertainty):
            self.x, self.variance \
                = err1d.pow2_inplace(self.x, self.variance, other.x, other.variance)
        else:
            self.x, self.variance = err1d.pow_inplace(self.x, self.variance, other)
        return self

    # always uses true division, even if division is not imported from future
    def __div__(self, other):
        return self.__truediv__(other)
    def __rdiv__(self, other):
        return self.__rtruediv__(other)
    def __idiv__(self, other):
        return self.__itruediv__(other)

    # Unary ops
    def __neg__(self):
        return Uncertainty(-self.x, self.variance)
    def __pos__(self):
        return self
    def __abs__(self):
        return Uncertainty(np.abs(self.x), self.variance)

    def __str__(self):
        #return str(self.x)+" +/- "+str(numpy.sqrt(self.variance))
        if np.isscalar(self.x):
            return format_uncertainty(self.x, np.sqrt(self.variance))
        else:
            content = ", ".join(format_uncertainty(v, dv)
                                for v, dv in zip(self.x, np.sqrt(self.variance)))
            return "".join(("[", content, "]"))
    def __repr__(self):
        return "Uncertainty(%s,%s)"%(str(self.x), str(self.variance))

    # Not implemented
    def __matmul__(self, other): return NotImplemented
    def __rmatmul__(self, other): return NotImplemented
    def __imatmul__(self, other): return NotImplemented

    # Integer operations not defined
    def __floordiv__(self, other): return NotImplemented
    def __mod__(self, other): return NotImplemented
    def __divmod__(self, other): return NotImplemented
    def __lshift__(self, other): return NotImplemented
    def __rshift__(self, other): return NotImplemented
    def __and__(self, other): return NotImplemented
    def __xor__(self, other): return NotImplemented
    def __or__(self, other): return NotImplemented

    def __rfloordiv__(self, other): return NotImplemented
    def __rmod__(self, other): return NotImplemented
    def __rdivmod__(self, other): return NotImplemented
    def __rlshift__(self, other): return NotImplemented
    def __rrshift__(self, other): return NotImplemented
    def __rand__(self, other): return NotImplemented
    def __rxor__(self, other): return NotImplemented
    def __ror__(self, other): return NotImplemented

    def __ifloordiv__(self, other): return NotImplemented
    def __imod__(self, other): return NotImplemented
    def __ilshift__(self, other): return NotImplemented
    def __irshift__(self, other): return NotImplemented
    def __iand__(self, other): return NotImplemented
    def __ixor__(self, other): return NotImplemented
    def __ior__(self, other): return NotImplemented

    def __invert__(self): return NotImplemented  # For ~x
    def __complex__(self): return NotImplemented
    def __int__(self): return NotImplemented
    def __float__(self): return NotImplemented  # don't silently ignore variance
    def __long__(self): return NotImplemented
    def __oct__(self): return NotImplemented
    def __hex__(self): return NotImplemented
    def __index__(self): return NotImplemented
    def __coerce__(self): return NotImplemented

    # Want to support numpy ufunc handling so that np.sin(A)
    # returns Uncertainty(*err1d.sin(A)) when A is an uncertainty object.
    # numpy.sin(x) first calls np.asarray(x) to coerce the input into an
    # array-like shape, then runs through the ufunc on each element of x.
    # This process will not work if __array__ is defined.
    #def __array__(self): return NotImplemented

    def mean(self, axis=None, dtype=None, out=None, keepdims=False, biased=True):
        r"""
        Compute the uncertainty-weighted average of the dataset.

        If varX is estimated from the data, then use *biased=True* so that the
        estimated variance is scaled by the normalized $\chi^2$.  See the
        wikipedia page for the weighted arithmetic mean for details.

        See numpy.mean for details on the remaining parameters.
        """
        if out is not None:
            out = out.x, out.variance
        M, varM = err1d.mean(x.x, x.variance, *args, **kw)
        return Uncertainty(M, varM)

    def sum(self, axis=None, dtype=None, out=None, keepdims=False):
        r"""
        Return the sum of an uncertain dataset.

        See numpy.sum for details on the interface.
        """
        if out is not None:
            new_out = out.x, out.variance
        else:
            new_out = (None, None)
        M, varM = err1d.sum(self.x, self.variance, axis=axis,
                            dtype=dtype, out=new_out, keepdims=keepdims)
        return Uncertainty(M, varM) if out is None else new_out

    def cumsum(self, axis=None, dtype=None, out=None, keepdims=False):
        r"""
        Return the cumulative sum of an uncertain dataset.

        See numpy.cumsum for details on the interface.
        """
        if out is not None:
            new_out = out.x, out.variance
        else:
            new_out = (None, None)
        M, varM = err1d.cumsum(self.x, self.variance, axis=axis,
                               dtype=dtype, out=new_out, keepdims=keepdims)
        return Uncertainty(M, varM) if out is None else new_out

    def std(self, *args, **kw):
        raise TypeError("Use mean() to compute the mean and variance of uncertainty items")

    def var(self, *args, **kw):
        raise TypeError("Use mean() to compute the mean and variance of uncertainty items")

    def average(self, axis=None, weights=None):
        r"""
        Return the weighted average of a data set.

        Note that mean(A) is the same as average(A, weights=1)
        """
        # TODO: give difference between mean and average
        M, varM = err1d.average(self.x, self.variance, w, w.variance, axis=axis)
        return Uncertainty(M, varM)

    # TODO: comparisons returning p-value?

    def log(self):
        return Uncertainty(*err1d.log(self.x, self.variance))
    def exp(self):
        return Uncertainty(*err1d.exp(self.x, self.variance))
    def sin(self):
        return Uncertainty(*err1d.sin(self.x, self.variance))
    def cos(self):
        return Uncertainty(*err1d.cos(self.x, self.variance))
    def tan(self):
        return Uncertainty(*err1d.tan(self.x, self.variance))
    def arcsin(self):
        return Uncertainty(*err1d.arcsin(self.x, self.variance))
    def arccos(self):
        return Uncertainty(*err1d.arccos(self.x, self.variance))
    def arctan(self):
        return Uncertainty(*err1d.arctan(self.x, self.variance))
    def arctan2(self, other):
        Uncertainty(*err1d.arctan2(self.x, self.variance, other.x, other.variance))

# numpy.method(x) calls x.method() if x has a method attribute, so just use
# the names from the numpy namespace.
sum = np.sum
cumsum = np.cumsum
mean = np.mean
#average = np.average   ## np.average(x) does not call x.average()
log = np.log
exp = np.exp
sin = np.sin
cos = np.cos
tan = np.tan
arcsin = np.arcsin
arccos = np.arccos
arctan = np.arctan
arctan2 = np.arctan2

def average(x, axis=None, weights=None):
    r"""
    Return the weighted average of a data set.
    """
    return x.average(axis=axis, weights=weights)


def interp(x, xp, fp, left=None, right=None):
    """
    Linear interpolation of x into points (xk, yk +/- dyk).

    xp is assumed to be in ascending order.

    left is the uncertainty value to return for points before the range of xp,
    or None for the initial value, fp[0].

    right is the uncertainty value to return for points after the range of xp,
    or None for the final value, fp[-1].
    """
    if left is None: left = fp[0]
    if right is None: right = fp[-1]
    left = (left.x, left.variance)
    right = (right.x, right.variance)

    if isinstance(x, np.ndarray) and x.ndim > 0:
        F, varF = err1d.interp(x, xp, fp.x, fp.variance, left, right)
    else:
        F, varF = err1d.interp([x], xp, fp.x, fp.variance, left, right)
        F, varF = F[0], varF[0]
    return Uncertainty(F, varF)


def smooth(x, xp, fp, degree=2, span=5):
    """
    Windowed least squares smoothing.
    """
    if span > 2:
        y, dy = wsolve.smooth(x, xp, fp, fp.dx, degree=degree, span=span)
        return Uncertainty(y, dy**2)
    else:
        # TODO: smooth with extrapolate, but interp will not.
        return interp(x, xp, fp)


def test():
    a = Uncertainty(5, 3)
    b = Uncertainty(4, 2)

    # Scalar operations
    z = a+4
    assert z.x == 5+4 and z.variance == 3
    z = a-4
    assert z.x == 5-4 and z.variance == 3
    z = a*4
    assert z.x == 5*4 and z.variance == 3*4**2
    z = a/4
    assert z.x == 5./4 and z.variance == 3./4**2

    # Reverse scalar operations
    z = 4+a
    assert z.x == 4+5 and z.variance == 3
    z = 4-a
    assert z.x == 4-5 and z.variance == 3
    z = 4*a
    assert z.x == 4*5 and z.variance == 3*4**2
    z = 4/a
    assert z.x == 4./5 and abs(z.variance - 3./5**4 * 4**2) < 1e-15

    # Power operations
    z = a**2
    assert z.x == 5**2 and z.variance == 4*3*5**2
    z = a**1
    assert z.x == 5**1 and z.variance == 3
    z = a**0
    assert z.x == 5**0 and z.variance == 0
    z = a**-1
    assert z.x == 5**-1 and abs(z.variance - 3./5**4) < 1e-15

    # Binary operations
    z = a+b
    assert z.x == 5+4 and z.variance == 3+2
    z = a-b
    assert z.x == 5-4 and z.variance == 3+2
    z = a*b
    assert z.x == 5*4 and z.variance == (5**2*2 + 4**2*3)
    z = a/b
    assert z.x == 5./4 and abs(z.variance - (3./5**2 + 2./4**2)*(5./4)**2) < 1e-15

    # ===== Inplace operations =====
    # Scalar operations
    y = a+0; y += 4
    z = a+4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y -= 4
    z = a-4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y *= 4
    z = a*4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y /= 4
    z = a/4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15

    # Power operations
    y = a+0; y **= 4
    z = a**4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15

    # Binary operations
    y = a+0; y += b
    z = a+b
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y -= b
    z = a-b
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y *= b
    z = a*b
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y /= b
    z = a/b
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15


    # =============== vector operations ================
    # Slicing
    z = Uncertainty(np.array([1, 2, 3, 4, 5]), np.array([2, 1, 2, 3, 2]))
    assert z[2].x == 3 and z[2].variance == 2
    assert (z[2:4].x == [3, 4]).all()
    assert (z[2:4].variance == [2, 3]).all()
    z[2:4] = Uncertainty(np.array([8, 7]), np.array([4, 5]))
    assert z[2].x == 8 and z[2].variance == 4
    A = Uncertainty(np.array([a.x]*2), np.array([a.variance]*2))
    B = Uncertainty(np.array([b.x]*2), np.array([b.variance]*2))

    # TODO complete tests of copy and inplace operations for vectors and slices.

    # Binary operations
    z = A+B
    assert (z.x == 5+4).all() and (z.variance == 3+2).all()
    z = A-B
    assert (z.x == 5-4).all() and (z.variance == 3+2).all()
    z = A*B
    assert (z.x == 5*4).all() and (z.variance == (5**2*2 + 4**2*3)).all()
    z = A/B
    assert (z.x == 5./4).all()
    assert (abs(z.variance - (3./5**2 + 2./4**2)*(5./4)**2) < 1e-15).all()

    # printing; note that sqrt(3) ~ 1.7
    assert str(Uncertainty(5, 3)) == "5.0(17)"
    assert str(Uncertainty(15, 3)) == "15.0(17)"
    assert str(Uncertainty(151.23356, 0.324185**2)) == "151.23(32)"

    xp = np.array([2., 3.])
    fp = Uncertainty(np.array([3., 1.]), np.array([0.04, 0.16]))
    z = interp(2.4, xp, fp)
    assert abs(z.x-2.2) < 2e-15
    assert abs(z.variance-0.04) < 2e-15
    z = interp([0., 2., 2.4, 3., 4.], xp, fp)
    assert np.linalg.norm(z.x - [3, 3, 2.2, 1, 1]) < 2e-15
    assert np.linalg.norm(z.variance - [0.04, 0.04, 0.04, 0.16, 0.16]) < 2e-15
    xp = np.array([2., 3., 3., 4.])
    fp = Uncertainty(np.array([3., 1., 3., 1.]), np.array([0.04, 0.16, 0.04, 0.16]))
    z = interp([2.5, 3., 3.5], xp, fp)
    assert np.linalg.norm(z.x - [2, 3, 2]) < 2e-15
    assert np.linalg.norm(z.variance - [0.05, 0.04, 0.05]) < 2e-15

if __name__ == "__main__": test()



from __future__ import print_function

import sys
import os
import importlib
from pprint import pprint
import traceback
import json
try:
    from urllib.parse import urlencode
    import urllib.request as urllib2
except ImportError:
    from urllib import urlencode
    import urllib2

import dataflow
from dataflow.core import Template, lookup_instrument, _instrument_registry
from dataflow.cache import use_redis, use_diskcache, get_cache
from dataflow.calc import process_template
import dataflow.core as df

import dataflow.modules

from dataflow import fetch

try:
    from reflweb import config
except ImportError:
    from reflweb import default_config as config

IS_PY3 = sys.version_info[0] >= 3

# for local and development installs of the server, the .git folder
# will exist in parent (reduction) folder...
if os.path.exists(os.path.join(os.path.dirname(__file__), "..", ".git")):
    import subprocess
    cwd = os.path.dirname(__file__)
    server_git_hash = subprocess.Popen(["git", "rev-parse", "HEAD"], cwd=cwd, stdout=subprocess.PIPE).stdout.read().strip()
    if IS_PY3: server_git_hash = server_git_hash.decode('ascii')
    server_mtime = int(subprocess.Popen(["git", "log", "-1", "--pretty=format:%ct"], cwd=cwd, stdout=subprocess.PIPE).stdout.read().strip())
    print("running git rev-parse HEAD", server_git_hash, server_mtime)
else:
    # otherwise for prebuilt systems use the packaged file that was created in setup.py
    import pkg_resources
    if pkg_resources.resource_exists("reflweb", "git_version_hash"):
        server_git_hash = pkg_resources.resource_string("reflweb", "git_version_hash").strip()
        if IS_PY3: server_git_hash = server_git_hash.decode('ascii')
        server_mtime = int(pkg_resources.resource_string("reflweb", "git_version_mtime").strip())
    else:
        server_git_hash = "unknown"
        server_mtime = 0

api_methods = []

def expose(action):
    """
    Decorator which adds function to the list of methods to expose in the api.
    """
    api_methods.append(action.__name__)
    return action

def sorted_ls(path, show_hidden=False):
    mtime = lambda f: os.stat(os.path.join(path, f)).st_mtime
    return list(sorted(filter(lambda x: os.path.exists(os.path.join(path, x)) and not x.startswith("."), os.listdir(path)), key=mtime))

def local_file_metadata(pathlist):
    # only absolute paths are supported:
    path = os.path.join(os.sep, *pathlist)
    dirlisting = sorted_ls(path)
    subdirs = []
    files = []
    files_metadata = {}
    for di in dirlisting:
        d = os.path.join(path, di)
        if os.path.isdir(d):
            subdirs.append(di)
        elif os.path.isfile(d):
            files.append(di)
            files_metadata[di] = {"mtime": int(os.path.getmtime(d))}
        else:
            # you've probably hit an unfulfilled path link or something.
            pass

    metadata = {
        "subdirs": subdirs,
        "files": files,
        "pathlist": pathlist,
        "files_metadata": files_metadata
        }
    return metadata

@expose
def get_file_metadata(source="ncnr", pathlist=None):
    if pathlist is None:
        pathlist = []

    if source == "local":
        metadata = local_file_metadata(pathlist)
    else:
        url = config.file_helper_url[source] #'http://ncnr.nist.gov/ipeek/listftpfiles.php'
        values = {'pathlist[]' : pathlist}
        data = urlencode(values, True)
        req = urllib2.Request(url, data.encode('ascii'))
        #print("request", url, data, req, type(req))
        response = urllib2.urlopen(req)
        fn = response.read()
        #print("response", fn)
        metadata = json.loads(fn.decode('ascii'))
        #print("parsed response", metadata)
        # this converts json to python object, then the json-rpc lib converts it
        # right back, but it is more consistent for the client this way:

    return metadata

@expose
def get_instrument(instrument_id="ncnr.refl"):
    """
    Make the instrument definition available to clients
    """
    instrument = lookup_instrument(instrument_id)
    return instrument.get_definition()

def refl_load(file_descriptors):
    """
    file_descriptors will be a list of dicts like
    [{"path": "ncnrdata/cgd/201511/21066/data/HMDSO_17nm_dry14.nxz.cgd", "mtime": 1447353278}, ...]
    """
    modules = [{"module": "ncnr.refl.load", "version": "0.1", "config": {}}]
    template = Template("test", "test template", modules, [], "ncnr.magik", version='0.0')
    retval = process_template(template, {0: {"files": file_descriptors}}, target=(0, "output"))
    return retval.todict()

@expose
def find_calculated(template_def, config):
    """
    Returns a vector of true/false for each node in the template indicating
    whether that node value has been calculated yet.
    """
    template = Template(**template_def)
    retval = dataflow.calc.find_calculated(template, config)
    return retval

@expose
def calc_terminal(template_def, config, nodenum, terminal_id, return_type='full'):
    """ json-rpc wrapper for calc_single
    template_def =
    {"name": "template_name",
     "description": "template description!",
     "modules": ["list of modules"],
     "wires": ["list of wires"],
     "instrument": "facility.instrument_name",
     "version": "2.7.3"
    }

    where modules in list of modules above have structure:
    module =
    {"module": "facility.instrument_name.module_name",
     "version": "0.3.2"
    }

    and wires have structure:
    [["wire_start_module_id:wire_start_terminal_id", "wire_end_module_id:wire_end_terminal_id"],
     ["1:output", "2:input"],
     ["0:xslice", "3:input"]
    ]

    config =
    [{"param": "value"}, ...]

    nodenum is the module number from the template for which you wish to get the calculated value

    terminal_id is the id of the terminal for that module, that you want to get the value from
    (output terminals only).
    """
    template = Template(**template_def)
    #print "template_def:", template_def, "config:", config, "target:",nodenum,terminal_id
    #print "modules","\n".join(m for m in df._module_registry.keys())
    try:
        retval = process_template(template, config, target=(nodenum, terminal_id))
    except Exception:
        print("==== template ===="); pprint(template_def)
        print("==== config ===="); pprint(config)
        #traceback.print_exc()
        raise
    if return_type == 'full':
        return retval.todict()
    elif return_type == 'plottable':
        return retval.get_plottable()
    elif return_type == 'metadata':
        return retval.get_metadata()
    elif return_type == 'export':
        # inject git version hash into export data:
        to_export = retval.get_export()
        to_export["server_git_hash"] = server_git_hash
        to_export["server_mtime"] = server_mtime
        return to_export

    raise KeyError(return_type + " not a valid return_type (should be one of ['full', 'plottable', 'metadata', 'export'])")

@expose
def calc_template(template_def, config):
    """ json-rpc wrapper for process_template """
    template = Template(**template_def)
    #print "template_def:", template_def, "config:", config
    try:
        retvals = process_template(template, config, target=(None, None))
    except Exception:
        print("==== template ===="); pprint(template_def)
        print("==== config ===="); pprint(config)
        #traceback.print_exc()
        raise
    output = {}
    for rkey, rv in retvals.items():
        module_id, terminal_id = rkey
        module_key = str(module_id)
        output.setdefault(module_key, {})
        output[module_key][terminal_id] = rv.todict()
    return output

@expose
def list_datasources():
    return config.data_sources

@expose
def list_instruments():
    return list(_instrument_registry.keys())

def create_instruments():

    fetch.DATA_SOURCES = config.data_sources

    if getattr(config, 'use_redis', False):
        redis_params = getattr(config, "redis_params", {})
        use_redis(**redis_params)
    elif getattr(config, 'use_diskcache', False):
        diskcache_params = getattr(config, "diskcache_params", {})
        use_diskcache(**diskcache_params)
    
    if getattr(config, 'use_compression', False):
        cache = get_cache()
        cache._use_compression = True

    # load refl instrument if nothing specified in config
    instruments = getattr(config, 'instruments', ['refl'])
    for instrument_id in instruments:
        instrument_module = importlib.import_module('dataflow.modules.{instr}'.format(instr=instrument_id))
        instrument_module.define_instrument()

if __name__ == '__main__':
    create_instruments()



import numpy as np
from .resolution import divergence

def apply_theta_offset(data, offset):
    data.sample.angle_x += offset
    data.detector.angle_x -= offset


def apply_absolute_angle(data):
    index = (data.sample.angle_x < 0)  # type: np.ndarray
    if np.any(index):
        data.sample.angle_x[index] *= -1.0
        data.detector.angle_x[index] *= -1.0


def apply_back_reflection(data):
    data.sample.angle_x *= -1.0
    data.detector.angle_x *= -1.0


def apply_divergence(data, sample_width, sample_broadening):
    distance = abs(data.slit1.distance), abs(data.slit2.distance)
    # for slit scans, the sample size and angle are irrelevant and
    # need to be excluded from the calculation of divergence:
    use_sample = data.intent != "intensity"
    if sample_width is None:
        sample_width = data.sample.width

    slits = data.slit1.x, data.slit2.x
    theta = data.sample.angle_x
    dtheta = divergence(slits=slits, distance=distance, T=theta,
                        sample_width=sample_width,
                        sample_broadening=sample_broadening,
                        use_sample=use_sample)
    #print "divergence",theta,slits,distance,sample_width,sample_broadening,dtheta
    data.angular_resolution = dtheta



"""
Manage the available templates for all instruments.
"""

# TODO: split instrument-specific files out of the dataflow package

def get_templates(instrument=''):
    """ 
    Returns a ist of pairs [name, template] for the given instrument.

    Templates are defined as JSON objects, with name stored in
    "reduction/dataflow/templates/<instrument>.<name>.json".
    """
    import os, json
    template_path = os.path.dirname(__file__)
    template_names = [fn
                      for fn in os.listdir(template_path)
                      if fn.endswith(".json") and fn.startswith(instrument)]
    templates = dict([(tn[len(instrument)+1:-5],
                       json.loads(open(os.path.join(template_path, tn), 'r').read()))
                      for tn in template_names])
    return templates




"""
To serve with uwsgi:

uwsgi --http 0.0.0.0:8002 --manage-script-name --mount /=server_flask:app

To serve with python:

python server_flask.py 8002
(then visit http://localhost:8002/static/index.html in your browser)

"""

import os, sys, posixpath
import traceback
import logging
import pkg_resources

from flask import Flask, request, make_response, redirect, send_from_directory
from werkzeug.exceptions import HTTPException
import msgpack as msgpack_converter

try:
    from reflweb import config
except ImportError:
    from reflweb import default_config as config

RPC_ENDPOINT = '/RPC2'
STATIC_PATH = pkg_resources.resource_filename('reflweb', 'static/')

app = Flask(__name__, static_folder=STATIC_PATH, static_url_path='/static')

@app.route('/')
def root():
    return redirect("static/index.html")

@app.route('/robots.txt')
def static_from_root():
    return send_from_directory(app.static_folder, request.path[1:])

@app.errorhandler(Exception)
def handle_error(e):
    code = 500
    if isinstance(e, HTTPException):
        code = e.code
    content = {'exception': repr(e), 'traceback': traceback.format_exc()}
    logging.info(content['traceback'])
    return make_response(msgpack_converter.dumps(content), code)

def wrap_method(mfunc):
    def wrapper(*args, **kwargs):
        real_kwargs = request.get_json() if request.get_data() else {}
        content = mfunc(*args, **real_kwargs)
        response = make_response(msgpack_converter.dumps(content))
        response.headers['Content-Type'] = 'application/msgpack'
        return response
    return wrapper

from reflweb import api
api.create_instruments()
for method in api.api_methods:
    mfunc = getattr(api, method)
    wrapped = wrap_method(mfunc)
    path = posixpath.join(RPC_ENDPOINT, method)
    shortpath = posixpath.join("/", method)
    app.add_url_rule(path, path, wrapped, methods=["POST"])
    app.add_url_rule(shortpath, shortpath, wrapped, methods=["POST"])

if __name__ == '__main__':
    logging.basicConfig(level=logging.WARNING)
    port = 8002
    if len(sys.argv) > 1:
        port = int(sys.argv[1])
    app.run(port=port)



#############################################################
# rename or copy this file to config.py if you make changes #
#############################################################

# change this to your fully-qualified domain name to run a
# remote server.  The default value of localhost will
# only allow connections from the same computer.
#
# for remote (cloud) deployments, it is advised to remove
# the "local" data_sources item below, and to serve static
# files using a standard webserver
#
# if use_redis is False, server will use in-memory cache.

# TODO: Convert this to JSON file in web-accesible ('static')
# directory.

jsonrpc_servername = "localhost"
jsonrpc_port = 8001
# ssl_args for https serving the rpc
# ssl_args = {"keyfile": None, "certfile": None}
http_port = 8000
serve_staticfiles = True
#use_redis = True
use_diskcache = True
diskcache_params = {"size_limit": int(4*2**30), "shards": 5}
data_sources = [
    {
        "name": "local",
        "url": "file:///",
        "start_path": "",
    },
    {
        "name": "ncnr",
        "url": "http://ncnr.nist.gov/pub/",
        "start_path": "ncnrdata",
    },
    {
        "name": "charlotte",
        "url": "http://charlotte.ncnr.nist.gov/pub",
        "start_path": "",
    },
    # set start_path for local files to usr/local/nice/server_data/experiments
    # for instrument installs
    #{
    #    "name": "ncnr_DOI",
    #    "DOI": "10.18434/T4201B",
    #},
]
file_helper_url = {
    "charlotte": "http://charlotte.ncnr.nist.gov/ipeek/listftpfiles.php",
    "ncnr": "https://ncnr.nist.gov/ipeek/listftpfiles.php",
    "ncnr_DOI": "https://ncnr.nist.gov/ipeek/listncnrdatafiles.php"
}
instruments = ["refl", "ospec", "sans"]



"""
SANS reduction steps
====================

Set of reduction steps for SANS reduction.
"""

from __future__ import print_function

from posixpath import basename, join
from copy import copy, deepcopy
from io import BytesIO

import numpy as np

from dataflow.lib.uncertainty import Uncertainty
from dataflow.lib import uncertainty

from .sansdata import SansData, Sans1dData, Parameters

ALL_ACTIONS = []
IGNORE_CORNER_PIXELS = True

def cache(action):
    """
    Decorator which adds the *cached* attribute to the function.

    Use *@cache* to force caching to always occur (for example, when
    the function references remote resources, vastly reduces memory, or is
    expensive to compute.  Use *@nocache* when debugging a function
    so that it will be recomputed each time regardless of whether or not it
    is seen again.
    """
    action.cached = True
    return action

def nocache(action):
    """
    Decorator which adds the *cached* attribute to the function.

    Use *@cache* to force caching to always occur (for example, when
    the function references remote resources, vastly reduces memory, or is
    expensive to compute.  Use *@nocache* when debugging a function
    so that it will be recomputed each time regardless of whether or not it
    is seen again.
    """
    action.cached = False
    return action

def module(action):
    """
    Decorator which records the action in *ALL_ACTIONS*.

    This just collects the action, it does not otherwise modify it.
    """
    ALL_ACTIONS.append(action)

    # This is a decorator, so return the original function
    return action

#################
# Loader stuff
#################

def url_load(fileinfo):
    from dataflow.fetch import url_get

    path, mtime, entries = fileinfo['path'], fileinfo['mtime'], fileinfo['entries']
    name = basename(path)
    fid = BytesIO(url_get(fileinfo))
    nx_entries = LoadMAGIKPSD.load_entries(name, fid, entries=entries)
    fid.close()
    return nx_entries

@cache
@module
def LoadSANS(filelist=None, flip=False, transpose=False, check_timestamps=True):
    """
    loads a data file into a SansData obj and returns that.
    Checks to see if data being loaded is 2D; if not, quits

    **Inputs**

    filelist (fileinfo[]): Files to open.

    flip (bool): flip the data up and down

    transpose (bool): transpose the data
    
    check_timestamps (bool): verify that timestamps on file match request

    **Returns**

    output (sans2d[]): all the entries loaded.

    2018-04-20 Brian Maranville
    """
    from dataflow.fetch import url_get
    from .loader import readSANSNexuz
    if filelist is None:
        filelist = []
    data = []
    for fileinfo in filelist:
        path, mtime, entries = fileinfo['path'], fileinfo.get('mtime', None), fileinfo.get('entries', None)
        name = basename(path)
        fid = BytesIO(url_get(fileinfo, mtime_check=check_timestamps))
        entries = readSANSNexuz(name, fid)
        for entry in entries:
            if flip:
                entry.data.x = np.fliplr(entry.data.x)
            if transpose:
                entry.data.x = entry.data.x.T
        data.extend(entries)

    return data

"""
    Variable vz_1 = 3.956e5		//velocity [cm/s] of 1 A neutron
	Variable g = 981.0				//gravity acceleration [cm/s^2]
	Variable m_h	= 252.8			// m/h [=] s/cm^2
////	//
	Variable yg_d,acc,sdd,ssd,lambda0,DL_L,sig_l
	Variable var_qlx,var_qly,var_ql,qx,qy,sig_perp,sig_para, sig_para_new

	G = 981.  //!	ACCELERATION OF GRAVITY, CM/SEC^2
	acc = vz_1 		//	3.956E5 //!	CONVERT WAVELENGTH TO VELOCITY CM/SEC
	SDD = L2		//1317
	SSD = L1		//1627 		//cm
	lambda0 = lambda		//		15
	DL_L = lambdaWidth		//0.236
	SIG_L = DL_L/sqrt(6)
	YG_d = -0.5*G*SDD*(SSD+SDD)*(LAMBDA0/acc)^2
/////	Print "DISTANCE BEAM FALLS DUE TO GRAVITY (CM) = ",YG
//		Print "Gravity q* = ",-2*pi/lambda0*2*yg_d/sdd

	sig_perp = kap*kap/12 * (3*(S1/L1)^2 + 3*(S2/LP)^2 + (proj_DDet/L2)^2)
	sig_perp = sqrt(sig_perp)


	FindQxQy(inQ,phi,qx,qy)

// missing a factor of 2 here, and the form is different than the paper, so re-write
//	VAR_QLY = SIG_L^2 * (QY+4*PI*YG_d/(2*SDD*LAMBDA0))^2
//	VAR_QLX = (SIG_L*QX)^2
//	VAR_QL = VAR_QLY + VAR_QLX  //! WAVELENGTH CONTRIBUTION TO VARIANCE
//	sig_para = (sig_perp^2 + VAR_QL)^0.5

	// r_dist is passed in, [=]cm
	// from the paper
	a_val = 0.5*G*SDD*(SSD+SDD)*m_h^2 * 1e-16		//units now are cm /(A^2)

    r_dist = sqrt(  (pixSize*((p+1)-xctr))^2 +  (pixSize*((q+1)-yctr)+(2)*yg_d)^2 )		//radial distance from ctr to pt

	var_QL = 1/6*(kap/SDD)^2*(DL_L)^2*(r_dist^2 - 4*r_dist*a_val*lambda0^2*sin(phi) + 4*a_val^2*lambda0^4)
	sig_para_new = (sig_perp^2 + VAR_QL)^0.5


///// return values PBR
	SigmaQX = sig_para_new
	SigmaQy = sig_perp

////

	results = "success"
	Return results
End
"""

#@cache
#@module
def calculateDQ(data):
    """
    Add the dQ column to the data, based on slit apertures and gravity
    r_dist is the real-space distance from ctr of detector to QxQy pixel location

    From `NCNR_Utils.ipf` (Steve R. Kline) in which the math is in turn from:

    | D.F.R Mildner, J.G. Barker & S.R. Kline J. Appl. Cryst. (2011). 44, 1127-1129.
    | *The effect of gravity on the resolution of small-angle neutron diffraction peaks*
    | [ doi:10.1107/S0021889811033322 ]

    **Inputs**

    data (sans2d): data in

    **Returns**

    output (sans2d): data in with dQ column filled in

    2017-06-16  Brian Maranville
    """

    G = 981.  #!    ACCELERATION OF GRAVITY, CM/SEC^2
    acc = vz_1 = 3.956e5 # velocity [cm/s] of 1 A neutron
    m_h	= 252.8			# m/h [=] s/cm^2
    # the detector pixel is square, so correct for phi
    DDetX = data.metadata["det.pixelsizex"]
    DDetY = data.metadata["det.pixelsizey"]
    x_offset = data.metadata["det.pixeloffsetx"]
    y_offset = data.metadata["det.pixeloffsety"]
    xctr = data.metadata["det.beamx"]
    yctr = data.metadata["det.beamy"]

    shape = data.data.x.shape
    x, y = np.indices(shape)
    X = DDetX * (x-xctr)
    Y = DDetY * (y-yctr)

    apOff = data.metadata["sample.position"]
    S1 = data.metadata["resolution.ap1"]
    S2 = data.metadata["resolution.ap2"]
    L1 = data.metadata["resolution.ap12dis"] - apOff
    L2 = data.metadata["det.dis"] + apOff
    LP = 1.0/( 1.0/L1 + 1.0/L2)
    SDD = L2
    SSD = L1
    lambda0 = data.metadata["resolution.lmda"]    #  15
    DL_L = data.metadata["resolution.dlmda"]    # 0.236
    YG_d = -0.5*G*SDD*(SSD+SDD)*(lambda0/acc)**2
    kap = 2.0*np.pi/lambda0
    phi = np.mod(np.arctan2(Y + 2.0*YG_d, X), 2.0*np.pi) # from x-axis, from 0 to 2PI
    proj_DDet = np.abs(DDetX*np.cos(phi)) + np.abs(DDetY*np.sin(phi))
    r_dist = np.sqrt(X**2 + (Y + 2.0*YG_d)**2)  #radial distance from ctr to pt

    sig_perp = kap*kap/12.0 * (3.0*(S1/L1)**2 + 3.0*(S2/LP)**2 + (proj_DDet/L2)**2)
    sig_perp = np.sqrt(sig_perp)

    a_val = 0.5*G*SDD*(SSD+SDD)*m_h**2 * 1e-16		# units now are cm /(A^2)

    var_QL = 1.0/6.0*((kap/SDD)**2)*(DL_L**2)*(r_dist**2 - 4.0*r_dist*a_val*(lambda0**2)*np.sin(phi) + 4.0*(a_val**2)*(lambda0**4))
    sig_para_new = np.sqrt(sig_perp**2 + var_QL)

    data.dq_perp = sig_perp
    data.dq_para = sig_para_new
    return data

@nocache
@module
def PixelsToQ(data, beam_center=[None,None], correct_solid_angle=True):
    """
    generate a q_map for sansdata. Each pixel will have 4 values: (qx, qy, q, theta)


    **Inputs**

    data (sans2d): data in

    beam_center {Beam Center Override} (coordinate?): If not blank, will override the beamx and beamy from the datafile.

    correct_solid_angle {Correct solid angle} (bool): Apply correction for mapping
        curved Ewald sphere to flat detector

    **Returns**

    output (sans2d): converted to I vs. Qx, Qy

    2016-04-16 Brian Maranville
    """

    L2 = data.metadata['det.dis']
    beamx_override, beamy_override = beam_center
    x0 = beamx_override if beamx_override is not None else data.metadata['det.beamx'] #should be close to 64
    y0 = beamy_override if beamy_override is not None else data.metadata['det.beamy'] #should be close to 64
    wavelength = data.metadata['resolution.lmda']
    shape = data.data.x.shape

    qx = np.empty(shape, 'float')
    qy = np.empty(shape, 'float')

    x, y = np.indices(shape) + 0.5 # left, bottom edge of first pixel is 0.5, 0.5 pix.
    X = data.metadata['det.pixelsizex']*(x-x0) # in mm in nexus, but converted by loader
    Y = data.metadata['det.pixelsizey']*(y-y0)
    r = np.sqrt(X**2+Y**2)
    theta = np.arctan2(r, L2)/2 #remember to convert L2 to cm from meters
    q = (4*np.pi/wavelength)*np.sin(theta)
    alpha = np.arctan2(Y, X)
    qx = q*np.cos(alpha)
    qy = q*np.sin(alpha)
    if correct_solid_angle:
        data.data.x = data.data.x * (np.cos(theta)**3)
    res = data.copy()
    #Adding res.q
    res.q = q
    res.qx = qx
    res.qy = qy
    res.metadata['det.beamx'] = x0
    res.metadata['det.beamy'] = y0
    q0 = (4*np.pi/wavelength)
    res.qx_min = q0/2.0 * data.metadata['det.pixelsizex']*(0.5 - x0)/ L2
    res.qy_min = q0/2.0 * data.metadata['det.pixelsizex']*(0.5 - y0)/ L2
    res.qx_max = q0/2.0 * data.metadata['det.pixelsizex']*(128.5 - x0)/ L2
    res.qy_max = q0/2.0 * data.metadata['det.pixelsizex']*(128.5 - y0)/ L2
    res.xlabel = "Qx (inv. Angstroms)"
    res.ylabel = "Qy (inv. Angstroms)"
    res.theta = theta
    return res

@cache
@module
def circular_av(data):
    """
    Using a circular average, it converts data to 1D (Q vs. I)


    **Inputs**

    data (sans2d): data in

    **Returns**

    nominal_output (sans1d): converted to I vs. nominal Q

    mean_output (sans1d): converted to I vs. mean Q within integrated region

    2016-04-11 Brian Maranville
    """
    from .draw_annulus_aa import annular_mask_antialiased

    #annular_mask_antialiased(shape, center, inner_radius, outer_radius,
    #                         background_value=0.0, mask_value=1.0, oversampling=8)

    # calculate the change in q that corresponds to a change in pixel of 1
    if data.qx is None:
        raise ValueError("Q is not defined - convert pixels to Q first")

    q_per_pixel = data.qx[1, 0]-data.qx[0, 0] / 1.0

    # for now, we'll make the q-bins have the same width as a single pixel
    step = q_per_pixel
    shape1 = data.data.x.shape
    x0 = data.metadata['det.beamx'] # should be close to 64
    y0 = data.metadata['det.beamy'] # should be close to 64
    L2 = data.metadata['det.dis']
    wavelength = data.metadata['resolution.lmda']

    center = (x0, y0)
    Qmax = data.q.max()
    Q = np.arange(step, Qmax, step) # start at first pixel out.
    Q_edges = np.zeros((Q.shape[0] + 1,), dtype="float")
    Q_edges[1:] = Q
    Q_edges += step/2.0 # get a range from step/2.0 to (Qmax + step/2.0)
    r_edges = L2 * np.tan(2.0*np.arcsin(Q_edges * wavelength/(4*np.pi))) / data.metadata['det.pixelsizex']
    Q_mean = []
    Q_mean_error = []
    I = []
    I_error = []
    dx = np.zeros_like(Q, dtype="float")
    for i, qq in enumerate(Q):
        # inner radius is the q we're at right now, converted to pixel dimensions:
        inner_r = r_edges[i]
        # outer radius is the q of the next bin, also converted to pixel dimensions:
        outer_r = r_edges[i+1]
        #print(i, qq, inner_r, outer_r)
        mask = annular_mask_antialiased(shape1, center, inner_r, outer_r)
        if IGNORE_CORNER_PIXELS:
            mask[0, 0] = mask[-1, 0] = mask[-1, -1] = mask[0, -1] = 0.0
        #print("Mask: ", mask)
        integrated_q = uncertainty.sum(data.q*mask.T)
        integrated_intensity = uncertainty.sum(data.data*mask.T)
        #error = getPoissonUncertainty(integrated_intensity)
        #error = np.sqrt(integrated_intensity)
        mask_sum = np.sum(mask)
        if mask_sum > 0.0:
            norm_integrated_intensity = integrated_intensity / mask_sum
            norm_integrated_q = integrated_q / mask_sum
            #error /= mask_sum
        else:
            norm_integrated_intensity = integrated_intensity
            norm_integrated_q = integrated_q

        I.append(norm_integrated_intensity.x) # not multiplying by step anymore
        I_error.append(norm_integrated_intensity.variance)
        Q_mean.append(norm_integrated_q)
        Q_mean_error.append(0.0)

    I = np.array(I, dtype="float")
    I_error = np.array(I_error, dtype="float")
    Q_mean = np.array(Q_mean, dtype="float")
    Q_mean_error = np.array(Q_mean_error, dtype="float")

    nominal_output = Sans1dData(Q, I, dx=dx, dv=I_error, xlabel="Q", vlabel="I",
                        xunits="inv. A", vunits="neutrons")
    nominal_output.metadata = deepcopy(data.metadata)
    nominal_output.metadata['extra_label'] = "_circ"

    mean_output = Sans1dData(Q_mean, I, dx=Q_mean_error, dv=I_error, xlabel="Q", vlabel="I",
                        xunits="inv. A", vunits="neutrons")
    mean_output.metadata = deepcopy(data.metadata)
    mean_output.metadata['extra_label'] = "_circ"

    return nominal_output, mean_output

@cache
@module
def sector_cut(data, angle=0.0, width=90.0, mirror=True):
    """
    Using annular averging, it converts data to 1D (Q vs. I)
    over a particular angle range


    **Inputs**

    data (sans2d): data in

    angle (float): center angle of sector cut (degrees)

    width (float): width of cut (degrees)

    mirror (bool): extend sector cut on both sides of origin
        (when false, integrates over a single cone centered at angle)

    **Returns**

    nominal_output (sans1d): converted to I vs. nominal Q

    mean_output (sans1d): converted to I vs. mean Q within integrated region

    2016-04-15 Brian Maranville
    """
    from .draw_annulus_aa import sector_cut_antialiased

    #annular_mask_antialiased(shape, center, inner_radius, outer_radius,
    #                         background_value=0.0, mask_value=1.0, oversampling=8)

    # calculate the change in q that corresponds to a change in pixel of 1
    q_per_pixel = data.qx[1, 0]-data.qx[0, 0] / 1.0

    # for now, we'll make the q-bins have the same width as a single pixel
    step = q_per_pixel
    shape1 = data.data.x.shape
    x0 = data.metadata['det.beamx'] # should be close to 64
    y0 = data.metadata['det.beamy'] # should be close to 64
    L2 = data.metadata['det.dis']
    wavelength = data.metadata['resolution.lmda']

    center = (x0, y0)
    Qmax = data.q.max()
    Q = np.arange(step, Qmax, step) # start at first pixel out.
    Q_edges = np.zeros((Q.shape[0] + 1,), dtype="float")
    Q_edges[1:] = Q
    Q_edges += step/2.0 # get a range from step/2.0 to (Qmax + step/2.0)
    r_edges = L2 * np.tan(2.0*np.arcsin(Q_edges * wavelength/(4*np.pi))) / data.metadata['det.pixelsizex']
    Q_mean = []
    Q_mean_error = []
    I = []
    I_error = []
    dx = np.zeros_like(Q, dtype="float")
    start_angle = np.radians(angle - width/2.0)
    end_angle = np.radians(angle + width/2.0)
    for i, qq in enumerate(Q):
        # inner radius is the q we're at right now, converted to pixel dimensions:
        inner_r = r_edges[i]
        # outer radius is the q of the next bin, also converted to pixel dimensions:
        outer_r = r_edges[i+1]
        #print(i, qq, inner_r, outer_r)
        mask = sector_cut_antialiased(shape1, center, inner_r, outer_r, start_angle=start_angle, end_angle=end_angle, mirror=mirror)
        if IGNORE_CORNER_PIXELS:
            mask[0, 0] = mask[-1, 0] = mask[-1, -1] = mask[0, -1] = 0.0
        #print("Mask: ", mask)
        integrated_q = uncertainty.sum(data.q*mask.T)
        integrated_intensity = uncertainty.sum(data.data*mask.T)
        #error = getPoissonUncertainty(integrated_intensity)
        #error = np.sqrt(integrated_intensity)
        mask_sum = np.sum(mask)
        if mask_sum > 0.0:
            norm_integrated_intensity = integrated_intensity / mask_sum
            norm_integrated_q = integrated_q / mask_sum
            #error /= mask_sum
        else:
            norm_integrated_intensity = integrated_intensity
            norm_integrated_q = integrated_q

        I.append(norm_integrated_intensity.x) # not multiplying by step anymore
        I_error.append(norm_integrated_intensity.variance)
        Q_mean.append(norm_integrated_q)
        Q_mean_error.append(0.0)

    I = np.array(I, dtype="float")
    I_error = np.array(I_error, dtype="float")
    Q_mean = np.array(Q_mean, dtype="float")
    Q_mean_error = np.array(Q_mean_error, dtype="float")

    nominal_output = Sans1dData(Q, I, dx=dx, dv=I_error, xlabel="Q", vlabel="I",
                        xunits="inv. A", vunits="neutrons")
    nominal_output.metadata = deepcopy(data.metadata)
    nominal_output.metadata['extra_label'] = "_%.1f" % (angle,)

    mean_output = Sans1dData(Q_mean, I, dx=Q_mean_error, dv=I_error, xlabel="Q", vlabel="I",
                        xunits="inv. A", vunits="neutrons")
    mean_output.metadata = deepcopy(data.metadata)
    mean_output.metadata['extra_label'] = "_%.1f" % (angle,)

    return nominal_output, mean_output

@module
def correct_detector_efficiency(sansdata):
    """
    Given a SansData object, corrects for the efficiency of the detection process

    **Inputs**

    sansdata (sans2d): data in

    **Returns**

    output (sans2d): corrected for efficiency

    2016-08-04 Brian Maranville and Andrew Jackson
    """

    L2 = sansdata.metadata['det.dis']
    lambd = sansdata.metadata["resolution.lmda"]
    shape = sansdata.data.x.shape
    (x0, y0) = np.shape(sansdata.data.x)
    x, y = np.indices(shape)
    X = sansdata.metadata['det.pixelsizex']/10.0*(x-x0/2)
    Y = sansdata.metadata['det.pixelsizey']/10.0*(y-y0/2)
    r = np.sqrt(X**2+Y**2)
    theta_det = np.arctan2(r, L2*100)/2

    stAl = 0.00967*lambd*0.8 # dimensionless, constants from JGB memo
    stHe = 0.146*lambd*2.5

    ff = (np.exp(-stAl/np.cos(theta_det))/np.exp(-stAl)
          * np.expm1(-stHe/np.cos(theta_det))/np.expm1(-stHe))

    res = sansdata.copy()
    res.data = res.data/ff

    # note that the theta calculated for this correction is based on the
    # center of the detector and NOT the center of the beam. Thus leave
    # the q-relevant theta alone.
    res.theta = copy(sansdata.theta)

    return res

@module
def correct_dead_time(sansdata, deadtime=1.0e-6):
    """
    Correct for the detector recovery time after each detected event
    (suppresses counts as count rate increases)

    **Inputs**

    sansdata (sans2d): data in

    deadtime (float): detector dead time (nonparalyzing?)

    **Returns**

    output (sans2d): corrected for dead time

    2010-01-03 Andrew Jackson?
    """

    dscale = 1.0/(1.0-deadtime*(np.sum(sansdata.data)/sansdata.metadata["run.rtime"]))

    result = sansdata.copy()
    result.data *= dscale
    return result

@module
def monitor_normalize(sansdata, mon0=1e8):
    """"
    Given a SansData object, normalize the data to the provided monitor

    **Inputs**

    sansdata (sans2d): data in

    mon0 (float): provided monitor

    **Returns**

    output (sans2d): corrected for dead time

    2010-01-01 Andrew Jackson?
    """
    monitor = sansdata.metadata['run.moncnt']
    res = sansdata.copy()
    res.data *= mon0/monitor
    return res

@module
def generate_transmission(in_beam, empty_beam, integration_box=[55, 74, 53, 72]):
    """
    To calculate the transmission, we integrate the intensity in a box
    for a measurement with the substance in the beam and with the substance
    out of the beam and take their ratio. The box is definied by xmin, xmax
    and ymin, ymax, I start counting at (0, 0).

    Coords are taken with reference to bottom left of the image.

    **Inputs**

    in_beam (sans2d): measurement with sample in the beam

    empty_beam (sans2d): measurement with no sample in the beam

    integration_box (range:xy): region over which to integrate

    **Returns**

    output (params): calculated transmission for the integration area

    2017-02-29 Brian Maranville
    """
    #I_in_beam = 0.0
    #I_empty_beam = 0.0
    #xmax, ymax = np.shape(in_beam.data.x)
    #print(xmax, ymax)
    # Vectorize this loop, it's quick, but could be quicker
    # test against this simple minded implementation
    #print(ymax-coords_bottom_left[1], ymax-coords_upper_right[1])

    #for x in range(coords_bottom_left[0], coords_upper_right[0]+1):
    #    for y in range(ymax-coords_upper_right[1], ymax-coords_bottom_left[1]+1):
    #        I_in_beam = I_in_beam+in_beam.data.x[x, y]
    #        I_empty_beam = I_empty_beam+empty_beam.data.x[x, y]

    xmin, xmax, ymin, ymax = map(int, integration_box)
    I_in_beam = np.sum(in_beam.data[xmin:xmax+1, ymin:ymax+1])
    I_empty_beam = np.sum(empty_beam.data[xmin:xmax+1, ymin:ymax+1])

    ratio = I_in_beam/I_empty_beam
    result = Parameters(factor=ratio.x, factor_variance=ratio.variance,
                        factor_err=np.sqrt(ratio.variance))

    return result

@module
def subtract(subtrahend, minuend):
    """
    Algebraic subtraction of datasets pixel by pixel

    **Inputs**

    subtrahend (sans2d): a in (a-b) = c

    minuend (sans2d?): b in (a-b) = c, defaults to zero

    **Returns**

    output (sans2d): c in (a-b) = c

    2010-01-01 unknown
    """
    if minuend is not None:
        return subtrahend - minuend
    else:
        return subtrahend

@module
def product(data, factor_param, propagate_error=True):
    """
    Algebraic multiplication of dataset

    **Inputs**

    data (sans2d): data in (a)

    factor_param (params?): multiplication factor (b), defaults to 1

    propagate_error {Propagate error} (bool): if factor_error is passed in, use it

    **Returns**

    output (sans2d): result (c in a*b = c)

    2010-01-02 unknown
    """
    if factor_param is not None:
        if propagate_error:
            variance = factor_param.get('factor_variance', 0.0)
        return data * Uncertainty(factor_param.get('factor', 1.0), variance)
    else:
        return data

@module
def divide(data, factor_param):
    """
    Algebraic multiplication of dataset

    **Inputs**

    data (sans2d): data in (a)

    factor_param (params): denominator factor (b), defaults to 1

    **Returns**

    output (sans2d): result (c in a*b = c)

    2010-01-01 unknown
    """
    if factor_param is not None:
        return data.__truediv__(factor_param['factor'])
    else:
        return data

def correct_solid_angle(sansdata):
    """
    Given a SansData with q, qx, qy, and theta images defined,
    correct for the fact that the detector is flat and the Ewald sphere
    is curved. Need to calculate theta first, so do PixelsToQ before this.

    **Inputs**

    data (sans2d): data in

    **Returns**

    output (sans2d): corrected for mapping to Ewald

    2016-08-03 Brian Maranville
    """

    sansdata.data.x = sansdata.data.x*(np.cos(sansdata.theta)**3)
    return sansdata

@cache
@module
def correct_detector_sensitivity(sansdata, sensitivity):
    """"
    Given a SansData object and an sensitivity map generated from a div,
    correct for the efficiency of the detector. Recall that sensitivities are
    generated by taking a measurement of plexiglass and dividing by the
    mean value

    **Inputs**

    sansdata (sans2d): data in (a)

    sensitivity (sans2d): data in (b)

    **Returns**

    output (sans2d): result c in a/b = c

    2017-01-04 unknown
    """
    res = sansdata.copy()
    res.data /= sensitivity.data

    return res

def lookup_attenuation(instrument_name, attenNo, wavelength):
    from .attenuation_constants import attenuation
    if attenNo == 0:
        return {"att": 1.0, "att_err": 0.0}

    ai = attenuation[instrument_name]
    attenNoStr = format(int(attenNo), 'd')
    att = ai['att'][attenNoStr]
    att_err = ai['att_err'][attenNoStr]
    wavelength_key = ai['lambda']

    wmin = np.min(wavelength_key)
    wmax = np.max(wavelength_key)
    if wavelength < wmin or wavelength > wmax:
        raise ValueError("Wavelength out of calibration range (%f, %f). You must manually enter the absolute parameters" % (wmin, wmax))

    w = np.array([wavelength], dtype="float")
    att_interp = np.interp(w, wavelength_key, att, 1.0, np.nan)
    att_err_interp = np.interp(w, wavelength_key, att_err)
    return {"att": att_interp[0], "att_err": att_err_interp[0]} # err here is percent error

@cache
@module
def correct_attenuation(sample, instrument="NG7"):
    """
    Divide by the attenuation factor from the lookup tables for the instrument

    **Inputs**

    sample (sans2d): measurement

    instrument (opt:NG7|NGB|NGB30): instrument name

    **Returns**

    atten_corrected (sans2d): corrected measurement
    """
    attenNo = sample.metadata['run.atten']
    wavelength = sample.metadata['resolution.lmda']
    attenuation = lookup_attenuation(instrument, attenNo, wavelength)
    att = attenuation['att']
    percent_err = attenuation['att_err']
    att_variance = (att*percent_err/100.0)**2
    denominator = Uncertainty(att, att_variance)
    atten_corrected = sample.copy()
    atten_corrected.attenuation_corrected = True
    atten_corrected.data /= denominator
    return atten_corrected

@cache
@module
def absolute_scaling(sample, empty, div, Tsam, instrument="NG7", integration_box=[55, 74, 53, 72]):
    """
    Calculate absolute scaling

    Coords are taken with reference to bottom left of the image.

    **Inputs**

    sample (sans2d): measurement with sample in the beam

    empty (sans2d): measurement with no sample in the beam

    div (sans2d): DIV measurement

    Tsam (params): sample transmission

    instrument (opt:NG7|NGB|NGB30): instrument name, should be NG7 or NG3

    integration_box (range:xy): region over which to integrate

    **Returns**

    abs (sans2d): data on absolute scale

    2017-01-13 Andrew Jackson
    """
    # data (that is going through reduction), empty beam,
    # div, Transmission of the sample, instrument(NG3.NG5, NG7)
    # ALL from metadata
    detCnt = empty.metadata['run.detcnt']
    countTime = empty.metadata['run.rtime']
    monCnt = empty.metadata['run.moncnt']
    sdd = empty.metadata['det.dis'] # already in cm
    pixel = empty.metadata['det.pixelsizex'] # already in cm
    lambd = wavelength = empty.metadata['resolution.lmda']

    if not empty.attenuation_corrected:
        attenNo = empty.metadata['run.atten']
        # Need attenTrans - AttenuationFactor - need to know whether NG3, NG5 or NG7 (acctStr)
        attenuation = lookup_attenuation(instrument, attenNo, wavelength)
        att = attenuation['att']
        percent_err = attenuation['att_err']
        att_variance = (att*percent_err/100.0)**2
        attenTrans = Uncertainty(att, att_variance)
    else:
        # If empty is already corrected for attenuation, don't do it here:
        attenTrans = Uncertainty(1.0, 0.0)

    #-------------------------------------------------------------------------------------#

    # Correct empty beam by the sensitivity
    data = empty.__truediv__(div.data)
    # Then take the sum in XY box, including stat. error
    xmin, xmax, ymin, ymax = map(int, integration_box)
    detCnt = np.sum(data.data[xmin:xmax+1, ymin:ymax+1])
    print("DETCNT: ", detCnt)

    #------End Result-------#
    # This assumes that the data is has not been normalized at all.
    # Thus either fix this or pass un-normalized data.
    # Compute kappa = incident intensity * solid angle of the pixel
    kappa = detCnt / attenTrans * 1.0e8 / monCnt * (pixel/sdd)**2
    #print("Kappa: ", kappa)

    #utc_datetime = date.datetime.utcnow()
    #print(utc_datetime.strftime("%Y-%m-%d %H:%M:%S"))

    Tsam_factor = Uncertainty(Tsam['factor'], Tsam['factor_variance'])

    #-----Using Kappa to Scale data-----#
    Dsam = sample.metadata['sample.thk']
    ABS = sample.__mul__(1/(kappa*Dsam*Tsam_factor))
    #------------------------------------
    return ABS

@cache
@module
def patchData(data1, data2, xmin=55, xmax=74, ymin=53, ymax=72):
    """
    Copies data from data2 to data1 within the defined patch region
    (often used for processing DIV files)

    **Inputs**

    data1 (sans2d): measurement to be patched

    data2 (sans2d): measurement to get the patch from

    xmin (int): left pixel of patch box

    xmax (int): right pixel of patch box

    ymin (int): bottom pixel of patch box

    ymax (int): top pixel of patch box

    **Returns**

    patched (sans2d): data1 with patch applied from data2

    """

    patch_slice = (slice(xmin, xmax+1), slice(ymin, ymax+1))
    output = data1.copy()
    output.data[patch_slice] = data2.data[patch_slice]
    return output

@cache
@module
def addSimple(data):
    """
    Naive addition of counts and monitor from different datasets,
    assuming all datasets were taken under identical conditions
    (except for count time)

    Just adds together count time, counts and monitor.

    Use metadata from first dataset for output.

    **Inputs**

    data (sans2d[]): measurements to be added together

    **Returns**

    sum (sans2d): sum of inputs

    2017-06-29  Brian Maranville
    """

    output = data[0].copy()
    for d in data[1:]:
        output.data += d.data
        output.metadata['run.moncnt'] += d.metadata['run.moncnt']
        output.metadata['run.rtime'] += d.metadata['run.rtime']
        output.metadata['run.detcnt'] += d.metadata['run.detcnt']
    return output


@cache
@module
def makeDIV(data1, data2, patchbox=(55, 74, 53, 72)):
    """
    Use data2 to patch the beamstop from data1 within the defined box, then
    divide by total counts and multiply by number of pixels.

    **Inputs**

    data1 (sans2d): base measurement (to be patched and normalized)

    data2 (sans2d): measurement to get the patch from

    patchbox (range:xy): box to apply the patch in

    **Returns**

    DIV (sans2d): data1 with patch applied from data2 and normalized

    2016-04-20 Brian Maranville
    """

    print("patchbox:", patchbox)
    xmin, xmax, ymin, ymax = map(int, patchbox)

    DIV = patchData(data1, data2, xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax)

    DIV.data = DIV.data / np.sum(DIV.data) * DIV.data.x.size

    return DIV

@module
def radialToCylindrical(data, theta_offset = 0.0, oversample_th = 2.0, oversample_r = 2.0):
    """
    Convert radial data to cylindrical coordinates

    **Inputs**

    data (sans2d): data to be transformed

    theta_offset (float): move the bounds of the output from the default (0 to 360 deg)

    oversample_th (float): oversampling in theta (to increase fidelity of output)

    oversample_r (float): oversampling in r

    **Returns**

    cylindrical (sans2d): transformed data

    mask (sans2d): normalization array

    2017-05-26 Brian Maranville
    """

    from .cylindrical import ConvertToCylindrical

    if data.qx is None or data.qy is None:
        xmin = -data.metadata['det.beamx']
        xmax = xmin + 128
        ymin = -data.metadata['det.beamy']
        ymax = ymin + 128
    else:
        xmin = data.qx.min()
        xmax = data.qx.max()
        ymin = data.qy.min()
        ymax = data.qy.max()

    print(xmin, xmax, ymin, ymax)
    _, normalization, normalized, extent = ConvertToCylindrical(data.data.x.T, xmin, xmax, ymin, ymax, theta_offset=theta_offset, oversample_th=oversample_th, oversample_r=oversample_r)

    output = data.copy()
    output.aspect_ratio = None
    output.data = Uncertainty(normalized.T, normalized.T)

    mask = data.copy()
    mask.aspect_ratio = None
    mask.data = Uncertainty(normalization.T, normalization.T)

    if data.qx is not None:
        output.qx = np.linspace(extent[0], extent[1], normalized.shape[1])
        # abusing the qx property here to mean "other x"
        mask.qx = output.qx.copy()
        output.xlabel = mask.xlabel = "theta (degrees)"

    if data.qy is not None:
        output.qy = np.linspace(extent[2], extent[3], normalized.shape[0])
        # abusing the qy property here to mean "other y"
        mask.qy = output.qy.copy()
        output.ylabel = mask.ylabel = "Q (inv. Angstrom)"

    return output, mask

@module
def sliceData(data, slicebox=[None,None,None,None]):
    """
    Sum 2d data along both axes and return 1d datasets

    **Inputs**

    data (sans2d) : data in
    
    slicebox (range?:xy): region over which to integrate (in data coordinates)

    **Returns**

    xout (sans1d) : xslice

    yout (sans1d) : yslice

    2018-04-20 Brian Maranville
    """
    
    if slicebox is None:
        slicebox = [None, None, None, None]
    xmin, xmax, ymin, ymax = slicebox
    
    res = data.copy()
    if data.qx is None or data.qy is None:
        # then use pixels
        xslice = slice(int(np.ceil(xmin)) if xmin is not None else None, int(np.floor(xmax)) if xmax is not None else None)
        yslice = slice(int(np.ceil(ymin)) if ymin is not None else None, int(np.floor(ymax)) if ymax is not None else None)
        x_in = np.arange(data.data.x.shape[0])
        y_in = np.arange(data.data.x.shape[1])
        x_out = x_in[xslice]
        y_out = y_in[yslice]
        dx = np.zeros_like(x_out)
        dy = np.zeros_like(y_out)
        
    else:
        # then use q-values
        qxmin = data.qx_min if data.qx_min is not None else data.qx.min()
        qxmax = data.qx_max if data.qx_max is not None else data.qx.max()
        qx_in = np.linspace(qxmin, qxmax, data.data.x.shape[0])
        qymin = data.qy_min if data.qy_min is not None else data.qy.min()
        qymax = data.qy_max if data.qy_max is not None else data.qy.max()
        qy_in = np.linspace(qymin, qymax, data.data.x.shape[1])
        
        xslice = slice(get_index(qx_in, xmin), get_index(qx_in, xmax))
        yslice = slice(get_index(qy_in, ymin), get_index(qy_in, ymax))
        x_out = qx_in[xslice]
        y_out = qy_in[yslice]
        dx = np.zeros_like(x_out)
        dy = np.zeros_like(y_out)
        
    dataslice = (xslice, yslice)
    x_sum = uncertainty.sum(data.data[dataslice], axis=1)
    y_sum = uncertainty.sum(data.data[dataslice], axis=0)
    
    x_output = Sans1dData(x_out, x_sum.x, dx=dx, dv=x_sum.variance, xlabel=data.xlabel, vlabel="I",
                    xunits="", vunits="neutrons", metadata=data.metadata)
    y_output = Sans1dData(y_out, y_sum.x, dx=dy, dv=y_sum.variance, xlabel=data.ylabel, vlabel="I",
                    xunits="", vunits="neutrons", metadata=data.metadata)
                        
    return x_output, y_output


@cache
@module
def SuperLoadSANS(filelist=None, do_det_eff=True, do_deadtime=True,
                  deadtime=1.0e-6, do_mon_norm=True, do_atten_correct=False, mon0=1e8,
                  check_timestamps=True):
    """
    loads a data file into a SansData obj, and performs common reduction steps
    Checks to see if data being loaded is 2D; if not, quits


    **Inputs**

    filelist (fileinfo[]): Files to open.

    do_det_eff {Detector efficiency corr.} (bool): correct detector efficiency

    do_deadtime {Dead time corr.} (bool): correct for detector efficiency drop due to detector dead time

    deadtime {Dead time value} (float): value of the dead time in the calculation above

    do_atten_correct {Attenuation correction} (bool): correct intensity for the attenuators in the beam

    do_mon_norm {Monitor normalization} (bool): normalize data to a provided monitor value

    mon0 (float): provided monitor
    
    check_timestamps (bool): verify that timestamps on file match request

    **Returns**

    output (sans2d[]): all the entries loaded.

    2018-04-20 Brian Maranville
    """
    data = LoadSANS(filelist, flip=False, transpose=False, check_timestamps=check_timestamps)

    if do_det_eff:
        data = [correct_detector_efficiency(d) for d in data]
    if do_deadtime:
        data = [correct_dead_time(d, deadtime=deadtime) for d in data]
    if do_mon_norm:
        data = [monitor_normalize(d, mon0=mon0) for d in data]
    if do_atten_correct:
        data = [correct_attenuation(d) for d in data]

    return data

def get_index(t, x):
    if (x == "" or x == None):
        return None
    if float(x) > t.max():
        return None
    if float(x) < t.min():
        return None
    tord = np.argsort(t)
    return tord[np.searchsorted(t, float(x), sorter=tord)]

def getPoissonUncertainty(y):
    """ for a poisson-distributed observable, get the range of
     expected actual values for a particular measured value.
     As described in the documentation for the error analysis
     on the BaBar experiment:

    4)      An alternative with some nice properties is +-0.5 + sqrt(n+0.25)
    i.e upper error = 0.5 + sqrt(n+0.25), lower error = -0.5 + sqrt(n+0.25).
    These produce the following intervals:
    n    low      high     cred.
    0 0.000000  1.000000 0.632121
    1 0.381966  2.618034 0.679295
    2 1.000000  4.000000 0.681595
    3 1.697224  5.302776 0.682159
    4 2.438447  6.561553 0.682378
    5 3.208712  7.791288 0.682485
    6 4.000000  9.000000 0.682545
    7 4.807418 10.192582 0.682582
    8 5.627719 11.372281 0.682607
    9 6.458619 12.541381 0.682624
    """
    hi = 0.5+np.sqrt(y+0.25)
    lo = -0.5+np.sqrt(y+0.25)
    #return {"yupper": y+hi, "ylower": y-lo, "hi": hi, "lo": lo}
    return {"yupper": y+hi, "ylower": y-lo}



from __future__ import print_function

import numpy as np

from dataflow.lib import err1d

def indent(text, prefix="  "):
    """
    Add a prefix to every line in a string.
    """
    return "\n".join(prefix+line for line in text.splitlines())


def group_data(datasets):
    """
    Groups data files by intent and polarization cross section.

    Returns a dictionary with the groups, keyed by (intent,polarization).
    """
    # TODO: also need to group by temperature/field
    groups = {}
    for d in datasets:
        groups.setdefault((d.intent, d.polarization), []).append(d)
    return groups


def group_by_xs(datasets):
    """
    Return datasets grouped by polarization cross section.
    """
    cross_sections = {}
    for data in datasets:
        cross_sections.setdefault(data.polarization, []).append(data)

    #print("datasets", [":".join((d.name, d.entry, d.polarization, d.intent)) for d in datasets])
    #print("xs", cross_sections)
    return cross_sections

def group_by_key(key, datasets):
    """
    Return datasets grouped by a value that can be found in a refldata file.
    Handle dotted namespace through recursive lookup.
    Handle union with comma. (e.g. key = "polarization,sample.name" would
    create group where sample.name and polarization are the same for all)
    """
    groups = {}
    key_items = key.split(",")
    for data in datasets:
        groupkey = []
        for item in key_items:
            item = item.strip()
            value = data
            for k in item.split("."):
                value = getattr(value, k)
            groupkey.append(value)
        groupkey = tuple(sorted(groupkey))
        groups.setdefault(groupkey, []).append(data)
    return groups


def group_by_intent(datasets):
    """
    Return datasets grouped by intent.
    """
    intents = {}
    for data in datasets:
        intents.setdefault(data.intent, []).append(data)

    #print("datasets", [":".join((d.name, d.entry, d.polarization, d.intent)) for d in datasets])
    #print("xs", cross_sections)
    return intents


def nearest(x, xp, fp=None):
    """
    Return the *fp* value corresponding to the *xp* that is nearest to *x*.

    If *fp* is missing, return the index of the nearest value.
    """
    if len(xp) == 1:
        if np.isscalar(x):
            return fp[0] if fp is not None else 0
        else:
            return np.array(len(x)*(fp if fp is not None else [0]))

    # if fp is not provided, want to return f as an index into the array xp
    # for the target values x, so set it to integer indices.  if fp is
    # provided, make sure it is an array.
    fp = np.arange(len(xp)) if fp is None else np.asarray(fp)


    # make sure that the xp array is sorted
    xp = np.asarray(xp)
    if np.any(np.diff(xp) < 0.):
        index = np.argsort(xp)
        xp, fp = xp[index], fp[index]

    # find the midpoints of xp and use that as the index
    xp = 0.5*(xp[:-1] + xp[1:])
    return fp[np.searchsorted(xp, x)]


def plot_sa(data):
    """
    Plot spin asymmetry data.
    """
    from matplotlib import pyplot as plt
    from uncertainties.unumpy import uarray as U, nominal_values, std_devs
    from dataflow.lib.errutil import interp
    # TODO: interp doesn't test for matching resolution
    data = dict((d.polarization, d) for d in data)
    pp, mm = data['++'], data['--']
    v_pp = U(pp.v, pp.dv)
    v_mm = interp(pp.x, mm.x, U(mm.v, mm.dv))
    sa = (v_pp - v_mm) / (v_pp + v_mm)
    v, dv = nominal_values(sa), std_devs(sa)
    plt.errorbar(pp.x, v, yerr=dv, fmt='.', label=pp.name)
    plt.xlabel("%s (%s)"%(pp.xlabel, pp.xunits) if pp.xunits else pp.xlabel)
    plt.ylabel(r'$(R^{++} -\, R^{--}) / (R^{++} +\, R^{--})$')


def test_nearest():
    # length 1 arrays
    xp, fp = [1], [5]
    assert nearest(0, xp) == 0
    assert (nearest([0], xp) == [0]).all()
    assert (nearest([0, 1], xp) == [0, 0]).all()
    assert nearest(0, xp, fp) == fp[0]
    assert (nearest([0], xp, fp) == [fp[0]]).all()
    assert (nearest([0, 1], xp, fp) == [fp[0]]*2).all()

    # constants as arrays
    xp, fp = [1, 1, 1], [5, 5, 5]
    assert nearest(0, xp) == 0
    assert (nearest([0], xp) == [0]).all()
    assert (nearest([0, 1], xp) == [0, 0]).all()
    assert nearest(0, xp, fp) == fp[0]
    assert (nearest([0], xp, fp) == [fp[0]]).all()
    assert (nearest([0, 1], xp, fp) == [fp[0]]*2).all()

    # actual arrays
    xp, fp = [1, 2, 3], [4, 5, 6]
    assert nearest(0, xp) == 0
    assert (nearest([0], xp) == [0]).all()
    assert (nearest([0, 1, 1.1, 1.6, 2.1, 2.9, 3, 3.1], xp)
            == [0, 0, 0, 1, 1, 2, 2, 2]).all()
    assert nearest(0, xp, fp) == fp[0]
    assert (nearest([0], xp, fp) == [fp[0]]).all()
    assert (nearest([0, 1, 1.1, 1.6, 2.1, 2.9, 3, 3.1], xp, fp)
            == [fp[i] for i in [0, 0, 0, 1, 1, 2, 2, 2]]).all()

    # unsorted arrays
    xp, fp = [1, 3, 2], [4, 5, 6]
    assert nearest(0, xp) == 0
    assert (nearest([0], xp) == [0]).all()
    assert (nearest([0, 1, 1.1, 1.6, 2.1, 2.9, 3, 3.1], xp)
            == [0, 0, 0, 2, 2, 1, 1, 1]).all()
    assert nearest(0, xp, fp) == fp[0]
    assert (nearest([0], xp, fp) == [fp[0]]).all()
    assert (nearest([0, 1, 1.1, 1.6, 2.1, 2.9, 3, 3.1], xp, fp)
            == [fp[i] for i in [0, 0, 0, 2, 2, 1, 1, 1]]).all()


def poisson_average(y, dy, norm='monitor'):
    r"""
    Return the Poisson average of a rate vector y +/- dy.

    Use *norm='monitor'* When counting against monitor (the default) or
    *norm='time'* when counting against time.

    The count rate is expressed as the number of counts in an interval $N$
    divided by the interval $M$.  The rate for the combined interval should
    match the rate you would get if you counted the entire interval at once,
    which is $\sum N_i / \sum M_i$.  We do this by inferring the counts
    and intervals from the rate and uncertainty, adding them together, and
    dividing to get the average rate over the entire interval.

    With counts and monitors both from Poisson distribution, the uncertainties
    are $\sqrt N$ and $\sqrt M$ respectively.  With error propagation, we get

    .. math::
       :nowrap:

        \begin{eqnarray}
        y &=& N/M \\
        \left(\frac{\Delta y}{y}\right)^2
            &=& \left(\frac{\Delta N}{N}\right)^2
              + \left(\frac{\Delta M}{M}\right)^2 \\
            &=& \left(\frac{1}{N}\right)
              + \left(\frac{1}{M}\right) \\
        \Delta y &=& y \sqrt{1/N + 1/M}
        \end{eqnarray}

    Inverting, we get

    .. math::
       :nowrap:

        \begin{eqnarray}
        M &=& y (y+1) / \Delta y^2 \\
        N &=& y M
        \end{eqnarray}

    When counts are zero, $y = 0$ and $M = 0$ according to the above formula.
    To correctly average intervals that may include zero counts, we need to
    be sure that the count uncertainty $\Delta N = 1$ when $N = 0$, which
    leads to a count rate of $0 \pm 1/M$, and the formula above needs to be
    adjusted accordingly, with $M = 1/ \Delta y$ when $N = 0$.

    To average a group of measurements $y_1, \ldots, y_n$ we first
    convert to counts and monitors, then set the rate to the sum of
    the counts over the sum of the monitors.  This gives

    .. math::
       :nowrap:

        \begin{eqnarray}
        M_i &=& y_i (y_i + 1) / \Delta y_i^2 \\
        N_i &=& y_i M_i \\
        \bar y &=& \sum N_i / \sum M_i \\
        \Delta \bar y &=& \bar y \sqrt{1/\sum_{N_i} + 1/\sum_{M_i}}
        \end{eqnarray}

    When counting against time the monitor uncertainty $\Delta M$ is
    effectively zero compared to uncertainty in the counts, and so
    the formulas are a little simpler:

    .. math::

        y &= N/M \\
        \Delta y &= y \sqrt{1/N} \\
        M &= y / \Delta y^2 \\
        N &= y M

    Again, zero counts leads to $M = 1 / \Delta y$.

    Averaging gives

    .. math::
       :nowrap:

        \begin{eqnarray}
        M_i &=& y_i / \Delta y_i^2 \\
        N_i &=& y_i M_i \\
        \bar y &=& \sum N_i / \sum M_i \\
        \Delta \bar y &=& \bar y \sqrt{1/\sum_{N_i}}
        \end{eqnarray}

    Our expression does not account for detector efficiency, attenuators
    and dead-time correction. In practice we have an additional scaling
    value $A \pm \Delta A$ for each point, giving

    .. math::
       :nowrap:

        \begin{eqnarray}
        y &=& A N/M \\
        \left(\frac{\Delta y}{y}\right)^2
            &=& \left(\frac{\Delta A}{A}\right)^2
              + \left(\frac{\Delta N}{N}\right)^2
              + \left(\frac{\Delta M}{M}\right)^2
        \end{eqnarray}

    Clearly with two inputs $y$, $\Delta y$ we cannot uniquely recover
    the four parameters $A$, $\Delta A$, $M$, $N$, and Poisson averaging
    will not work properly.  In these cases, a monitor weighted average
    is better for error propagation.  Monitor weighted averaging
    works well for everything except data with many zero counts.
    """
    dy = dy + (dy == 0)  # Protect against zero counts in division

    # Recover monitor and counts
    if norm == "monitor":
        monitors = y*(y+1)/dy**2
    else:
        monitors = y/dy**2
    monitors[y == 0] = 1./dy[y == 0]  # Special handling for 0 counts
    counts = y*monitors

    # Compute average rate
    combined_monitors, combined_counts = np.sum(monitors), np.sum(counts)
    bar_y = combined_counts/combined_monitors
    if norm == "monitor":
        if bar_y == 0:
            # then the simplification: sqrt(1/N + 1/M) is invalid,
            # use |dy| = 1/M*sqrt((dN)^2 + 1/M)
            # but dN in this case is 1.
            bar_dy = 1./combined_monitors * np.sqrt(1. + 1./combined_monitors)
        else:
            # the simplification will work in this case, sort of
            bar_dy = bar_y * np.sqrt(1./combined_counts + 1./combined_monitors)
    else:
        bar_dy = bar_y * np.sqrt(1./combined_monitors)

    #print("est. monitors:", monitors)
    #print("est. counts:", counts)
    return bar_y, bar_dy


def gaussian_average(y, dy, w, dw=0):
    bar_y, bar_y_var = err1d.average(y, dy**2, w, dw**2)
    return bar_y, np.sqrt(bar_y_var)


def demo_error_prop(title, rate, monitors, attenuators=None,
                    norm='monitor'):
    """
    Compare point averaging algorithms on simulated data.

    *title* is the label for the comparison.

    *rate* is the underlying count rate.

    *monitors* is the set of counting intervals, or counting time if
    Plugging in some numbers combining two counts, one for 2000 monitors
    and one for 4000 monitors, we can compute the difference between
    measuring the values separately and measuring them together.  Using
    count rates which are low relative to the monitor yields a relative
    error on the order of 1/10^6:

        Na=7, Ma=2000, Nb=13, Mb=4000, Aa=Ab=1, dAa=dAb=0

    When count rates are high, such as with a direct beam measurement where
    the monitor rate 10% of the detector rate, the relative error is on
    the order of 0.02%::

        Na=20400, Ma=2000, Nb=39500, Mb=4000

    Monitor uncertainty is significant at high count rates.  In the above
    example, the rate uncertainty dr/r for Na/Ma is found to be 2.4% when
    monitor uncertainty is included, but only 0.7% if monitor uncertainty is
    not included in the calculation.  At high Q, where uncertainty count rates
    are much lower than the monitor rate, monitor uncertainty is much less
    important.

    The effect of Poisson versus gaussian averaging is marginal, even for
    regions with extremely low counts, so long as the gaussian average is
    weighted by monitor counts.
    """
    from uncertainties import ufloat
    from uncertainties.unumpy import uarray, nominal_values as uval, \
        std_devs as udev
    from numpy.random import poisson

    #norm='time'
    #norm='monitor'
    time_err = 0
    #time_err = 0.1

    use_attenuators = attenuators is not None
    monitors = np.array(monitors, 'd')
    if norm == 'monitor':
        umonitors = uarray(monitors, np.sqrt(monitors))
    else:
        umonitors = uarray(monitors, time_err)
    if use_attenuators:
        uattenuators = uarray(*list(zip(*attenuators)))
    else:
        uattenuators = uarray(monitors/monitors, 0.*monitors)
    expected = rate*monitors/uval(uattenuators)
    # TODO: simulation is not correct for count by monitor
    counts = poisson(expected)    # Simulated counts
    #counts = expected   # Non-simulated counts
    #ucounts = uarray(counts, np.sqrt(counts + (counts==0)))
    #ucounts = uarray(counts + (counts==0), np.sqrt(counts + (counts==0)))
    ucounts = uarray(counts, np.sqrt(counts))
    incident = ucounts*uattenuators

    print("="*10, title+", rate=%g,"%rate, \
          "median counts=%d"%np.median(counts), "="*10)

    # rate averaged across different counting intervals
    y = incident/umonitors
    y_ave = poisson_average(uval(y), udev(y), norm=norm)
    y_gm = np.sum(monitors*y)/np.sum(monitors)
    y_ave = ufloat(*y_ave)
    y_g = np.mean(y)

    # rate estimated from full counting time

    tin = (np.sum(incident) if use_attenuators
           else ufloat(np.sum(counts), np.sqrt(np.sum(counts))))
    if norm == 'monitor':
        tmon = ufloat(np.sum(monitors), np.sqrt(np.sum(monitors)))
    else:
        tmon = ufloat(np.sum(monitors), len(monitors)*time_err)
    direct = tin/tmon

    #print("monitors:", monitors)
    #print("counts:", counts)
    #print("incident:", incident)
    if use_attenuators:
        print("attenuators:", list(zip(*attenuators))[0])
    def show(label, r, tag=""):
        if r is direct:
            rel = ""
        else:
            rel = (" diff: (%.1f,%.1f)%% "
                   % (100*(r.n-direct.n)/direct.n, 100*(r.s-direct.s)/direct.s))
        print(label, "r:", r, " dr/r: %.2f%%"%(100*r.s/r.n), rel, tag)
    show("Combined", direct)
    show("Poisson ", y_ave)
    show("Gaussian", y_g)
    show("Gaussian", y_gm, "monitor weighted")

    # again without monitor uncertainty
    if 0:
        y2 = incident/monitors
        y2_ave = ufloat(*poisson_average(uval(y2), udev(y2)))
        y2_g = np.mean(y2)
        show("Separate", y2_ave, "no monitor uncertainty")
        show("Gaussian", y2_g, "no monitor uncertainty")


def _fetch_url_uncached(url):
    # type: (str) -> str
    import urllib2
    try:
        fd = urllib2.urlopen(url)
        ret = fd.read()
    finally:
        fd.close()
    return ret


def fetch_url(url, url_cache="/tmp"):
    """
    Fetch a file from a url.

    The content of the file will be cached in */tmp/_path_to_file.ext* for
    the url *http://path/to/file.ext*.  If *url_cache* is None, then no caching
    is performed.  Set *url_cache* to the private cache directory if you don't
    want the files cleaned by the tmp reaper.
    """
    from urlparse import urlparse
    import os

    if url_cache is None:
        return _fetch_url_uncached(url)

    cached_path = os.path.join(url_cache, urlparse(url).path.replace('/', '_'))
    if os.path.exists(cached_path):
        with open(cached_path) as fd:
            ret = fd.read()
    else:
        ret = _fetch_url_uncached(url)
        with open(cached_path, 'wb') as fd:
            fd.write(ret)
    return ret


if __name__ == "__main__":
    demo_error_prop("mixed monitor", 10.0, [2000, 2000, 4000])
    demo_error_prop("mixed monitor", 1.0, [2000, 2000, 4000])
    demo_error_prop("mixed monitor", 0.035, [2000, 2000, 4000])
    demo_error_prop("mixed monitor", 0.0035, [2000, 2000, 4000])
    #demo_error_prop("mixed monitor", 0.00035, [2000, 2000, 4000])
    demo_error_prop("same monitor", 10.0, [2000]*700)
    demo_error_prop("same monitor", 1.0, [2000]*700)
    demo_error_prop("same monitor", 0.035, [2000]*700)
    demo_error_prop("same monitor", 0.0035, [2000]*700)
    demo_error_prop("same monitor", 0.00035, [2000]*700)
    demo_error_prop("attenuators", 10.0, [2000, 2000, 4000], [(1, 0), (12, 0.2), (12, 0.2)])
    demo_error_prop("attenuators", 10.0, [2000, 4000], [(12, 0.2), (122, 0.2)])



import datetime
from os.path import basename
try:
    import urllib.request as urllib2
except ImportError:
    import urllib2

import pytz

from dataflow.lib.iso8601 import seconds_since_epoch
from dataflow.fetch import url_get

from . import nexusref
from . import xrawref

DATA_SOURCES = {}

def check_datasource(source):
    if not source in DATA_SOURCES:
        raise RuntimeError("Need to set reflred.steps.load.DATA_SOURCES['" + source + "'] first!")

def url_load(fileinfo, check_timestamps=True):
    path, mtime, entries = fileinfo['path'], fileinfo.get('mtime', None), fileinfo.get('entries', None)
    filename = basename(path)
    content = url_get(fileinfo, mtime_check=check_timestamps)
    if filename.endswith('.raw') or filename.endswith('.ras'):
        return xrawref.load_from_string(filename, content, entries=entries)
    else:
        return nexusref.load_from_string(filename, content, entries=entries)

def find_mtime(path, source="ncnr"):
    check_datasource(source)
    try:
        url = urllib2.urlopen(DATA_SOURCES[source]+path)
        mtime = url.info().getdate('last-modified')
    except urllib2.HTTPError as exc:
        raise ValueError("Could not open %r\n%s"%(path, str(exc)))
    mtime_obj = datetime.datetime(*mtime[:7], tzinfo=pytz.utc)
    timestamp = seconds_since_epoch(mtime_obj)

    return {'path': path, 'mtime': timestamp, 'source': source, 'entries': None}


def url_load_list(files=None, check_timestamps=True):
    if files is None:
        return []
    result = [entry for fileinfo in files for entry in url_load(fileinfo, check_timestamps=check_timestamps)]
    return result



import numpy as np

def rebin_counts_broadcast(x, I, xo, Io, ND_portion=1.0):
    """
    Identify overlaps in bin edges and rebin
    Assumes both x and xo are sorted in increasing order.
    """
    # overlap map (MxN), where M is input size(x) and N is output size (xo):
    overlap = (np.minimum(xo[None, 1:], x[1:, None])
               - np.maximum(xo[None, :-1], x[:-1, None])).clip(min=0.0)
    portion = overlap / (x[1:] - x[:-1])[:, None]
    Io += np.sum(I[:, None] * ND_portion * portion, axis=0)

def rebin_counts_portion_transfer(x, I, xo, Io, ND_portion=1.0):
    """
    Identify overlaps in bin edges and rebin
    Assumes both x and xo are sorted in increasing order.
    """

    # overlap map (MxN), where M is input size(x) and N is output size (xo):
    overlap = (np.minimum(xo[None, 1:], x[1:, None])
               - np.maximum(xo[None, :-1], x[:-1, None])) #.clip(min=0.0)
    mask = (overlap > 0)
    xi, xoi = np.indices(overlap.shape)
    inlist = xi[mask]
    outlist = xoi[mask]
    weights = (overlap * I[:, None] * ND_portion / (x[1:] - x[:-1])[:, None])[mask]
    hist, bins = np.histogram(outlist, bins=np.arange(Io.shape[0]+1), weights=weights)
    Io += hist

def rebin_counts_portion(x, I, xo, Io, ND_portion=1.0):
    # use this one for 1D: very fast.
    cc = np.empty_like(x)
    cc[0] = 0.0
    cc[1:] = np.cumsum(I*ND_portion)
    integral = np.interp(xo, x, cc)
    Io += np.diff(integral)

def rebin_counts(x, I, xo, Io):
    input_direction = np.sign(x[-1] - x[0])
    output_direction = np.sign(xo[-1] - xo[0])
    input_slice = slice(None, None, int(input_direction))
    output_slice = slice(None, None, int(output_direction))

    Io[:] = 0.0
    rebin_counts_portion(x[input_slice], I[input_slice], xo[output_slice], Io[output_slice], ND_portion=1.0)

def rebin_counts_2D_bruteforce(x, y, I, xo, yo, Io):

    xoverlap = (np.minimum(xo[None, 1:], x[1:, None]) - np.maximum(xo[None, :-1], x[:-1, None])).clip(min=0.0)
    xportion = xoverlap / (x[1:] - x[:-1])[:, None]
    yoverlap = (np.minimum(yo[None, 1:], y[1:, None]) - np.maximum(yo[None, :-1], y[:-1, None])).clip(min=0.0)
    yportion = yoverlap / (y[1:] - y[:-1])[:, None]

    weighted_overlap = I[:, None, :, None] * xportion[:, :, None, None] * yportion[None, None, :, :]
    Io += np.sum(np.sum(weighted_overlap, axis=0), axis=1)

def rebin_counts_2D_indexing(x, y, I, xo, yo, Io):
    output = np.empty_like(Io)
    csx = np.empty((len(x), len(y)-1))
    ix = np.empty((len(xo)-1, len(y)-1))
    csx[0, :] = 0.0
    csx[1:, :] = np.cumsum(I, axis=0)
    for i in range(len(y)-1):
        integral = np.interp(xo, x, csx[:, i])
        # integral has shape (len(xo),)
        ix[:, i] = np.diff(integral)
        # ix[:,i] has shape (len(xo)-1,)
        # np.diff(integral) has shape(len(xo)-1,)

    csy = np.empty((len(xo)-1, len(y)))
    # shape is (len(xo)-1, len(y)-1)
    csy[:, 0] = 0.0
    csy[:, 1:] = np.cumsum(ix, axis=1)
    for j in range(len(xo)-1):
        integral = np.interp(yo, y, csy[j, :])
        # has shape (len(yo),)
        output[j, :] = np.diff(integral)
        # output[j, :] has shape (len(yo)-1)

    Io += output

def rebin_counts_2D_indexing_new(x, y, I, xo, yo, Io):
    # use this one for 2D.
    csx = np.empty((len(x), len(y)-1))
    csx[0, :] = 0.0
    csx[1:, :] = np.cumsum(I, axis=0)

    xindices = np.interp(xo, x, np.arange(len(x), dtype='float'))
    xindices_whole = np.floor(xindices).astype(int)
    xindices_frac = np.fmod(xindices, 1.0)
    # the only way an index will match the highest bin edge is for lookups at or outside the range of the
    # source.  In this case the fractional portion will always == 0.0, so clipping this to the bin edge below
    # is safe and allows us to use the source weights unmodified
    xintegral = csx[xindices_whole, :] + I[xindices_whole.clip(max=len(x)-2), :]*xindices_frac[:, None]
    # rebinned over x
    ix = np.diff(xintegral, axis=0)

    csy = np.empty((len(xo)-1, len(y)))
    csy[:, 0] = 0.0
    csy[:, 1:] = np.cumsum(ix, axis=1)

    yindices = np.interp(yo, y, np.arange(len(y), dtype='float'))
    yindices_whole = np.floor(yindices).astype(int)
    yindices_frac = np.fmod(yindices, 1.0)
    yintegral = csy[:, yindices_whole] + ix[:, yindices_whole.clip(max=len(y)-2)]*yindices_frac[None, :]
    # rebinned over x and y
    ixy = np.diff(yintegral, axis=1)

    Io += ixy

def rebin_counts_2D_indexing_newer(x, y, I, xo, yo, Io):
    # use this one for 2D.
    csx = np.empty((len(x), len(y)-1))
    csx[0, :] = 0.0
    csx[1:, :] = I
    np.cumsum(csx, out=csx, axis=0)

    xindices = np.interp(xo, x, np.arange(len(x), dtype='float'))
    xindices_whole = np.floor(xindices).astype(int)
    xindices_frac = np.fmod(xindices, 1.0)
    # the only way an index will match the highest bin edge is for lookups at or outside the range of the
    # source.  In this case the fractional portion will always == 0.0, so clipping this to the bin edge below
    # is safe and allows us to use the source weights unmodified
    xintegral = I[xindices_whole.clip(max=len(x)-2), :]
    xintegral *= xindices_frac[:, None]
    xintegral += csx[xindices_whole, :]
    del csx
    # rebinned over x
    ix = np.diff(xintegral, axis=0)
    # Note: inplace diff, which returns -delta in xintegral[0:-1], is only a
    # few percent faster, and is much more subject to implementation details
    # in the overlapping region, so don't use it.  Numpy 1.13 promises to
    # create a temporary for inplace operations on overlapping regions, so
    # the minor performance gains are likely to disappear anyway.
    #xintegral[:-1] -= xintegral[1:]
    #ix = xintegral[:-1]

    csy = np.empty((len(xo)-1, len(y)))
    csy[:, 0] = 0.0
    csy[:, 1:] = ix
    np.cumsum(csy, out=csy, axis=1)

    yindices = np.interp(yo, y, np.arange(len(y), dtype='float'))
    yindices_whole = np.floor(yindices).astype(int)
    yindices_frac = np.fmod(yindices, 1.0)
    yintegral = ix[:, yindices_whole.clip(max=len(y)-2)]
    yintegral *= yindices_frac[None, :]
    yintegral += csy[:, yindices_whole]
    del csy
    # rebinned over x and y
    ixy = np.diff(yintegral, axis=1)
    #yintegral[:-1] -= yintegral[1:]
    #ixy = yintegral[:-1]

    Io += ixy

def rebin_counts_2D(x, y, I, xo, yo, Io):
    x_input_direction = np.sign(x[-1] - x[0])
    y_input_direction = np.sign(y[-1] - y[0])
    x_output_direction = np.sign(xo[-1] - xo[0])
    y_output_direction = np.sign(yo[-1] - yo[0])

    x_s = slice(None, None, int(x_input_direction))
    y_s = slice(None, None, int(y_input_direction))
    xo_s = slice(None, None, int(x_output_direction))
    yo_s = slice(None, None, int(y_output_direction))

    Io[:] = 0
    rebin_counts_2D_indexing_newer(x[x_s], y[y_s], I[x_s, y_s], xo[xo_s], yo[yo_s], Io[xo_s, yo_s])

def test(sizex=101):
    sizexo = int(sizex/2)
    x = np.linspace(0.0, 100.0, sizex)
    xo = np.linspace(0.0, 110.0, sizexo)
    I = np.ones((sizex-1,), dtype="float")
    Io = np.zeros((sizexo-1,), dtype="float")

    rebin_counts_portion(x, I, xo, Io)
    return Io

def test_c(sizex=101):
    from . import rebin
    sizexo = int(sizex/2)
    x = np.linspace(0.0, 100.0, sizex)
    xo = np.linspace(0.0, 110.0, sizexo)
    I = np.ones((sizex-1,), dtype="float")
    Io = np.zeros((sizexo-1,), dtype="float")

    rebin.rebin(x, I, xo, Io)
    return Io

def test2D(sizex=101, sizey=75):
    sizexo = int(sizex/2)
    sizeyo = int(sizey/2)
    x = np.linspace(5.0, 100.0, sizex)
    xo = np.linspace(0.0, 110.0, sizexo)
    y = np.linspace(5.0, 50.0, sizey)
    yo = np.linspace(0.0, 60.0, sizeyo)
    I = np.ones((sizex-1, sizey-1), dtype="float")
    Io = np.zeros((sizexo-1, sizeyo-1), dtype="float")

    rebin_counts_2D_indexing_new(x, y, I, xo, yo, Io)
    return I, Io

def test2D_newer(sizex=101, sizey=75):
    sizexo = int(sizex/2)
    sizeyo = int(sizey/2)
    x = np.linspace(5.0, 100.0, sizex)
    xo = np.linspace(0.0, 110.0, sizexo)
    y = np.linspace(5.0, 50.0, sizey)
    yo = np.linspace(0.0, 60.0, sizeyo)
    I = np.ones((sizex-1, sizey-1), dtype="float")
    Io = np.zeros((sizexo-1, sizeyo-1), dtype="float")

    rebin_counts_2D_indexing_newer(x, y, I, xo, yo, Io)
    return I, Io

def test2D_c(sizex=101, sizey=75):
    from . import rebin
    sizexo = int(sizex/2)
    sizeyo = int(sizey/2)
    x = np.linspace(5.0, 100.0, sizex)
    xo = np.linspace(0.0, 110.0, sizexo)
    y = np.linspace(5.0, 50.0, sizey)
    yo = np.linspace(0.0, 60.0, sizeyo)
    I = np.ones((sizex-1, sizey-1), dtype="float")
    Io = np.zeros((sizexo-1, sizeyo-1), dtype="float")

    Io = rebin.rebin2d(x, y, I, xo, yo)
    return I, Io



#!/usr/bin/env python
import sys, os
from os.path import join as joinpath, dirname
import re

if len(sys.argv) == 1:
    sys.argv.append('install')

# Use our own nose-based test harness
if sys.argv[1] == 'test':
    from subprocess import call
    sys.exit(call([sys.executable, 'test.py'] + sys.argv[2:]))

#sys.dont_write_bytecode = True

from setuptools import setup, Extension, find_packages

import subprocess
git_version_hash = subprocess.Popen(["git", "rev-parse", "HEAD"], stdout=subprocess.PIPE).stdout.read()
open("reflweb/git_version_hash", "wb").write(git_version_hash)
server_mtime = subprocess.Popen(["git", "log", "-1", "--pretty=format:%ct"], stdout=subprocess.PIPE).stdout.read()
open("reflweb/git_version_mtime", "wb").write(server_mtime)

packages = find_packages(exclude=['reflbin'])

#sys.dont_write_bytecode = False
dist = setup(
    name='reductus',
    version='0.1b2',
    author='Paul Kienzle',
    author_email='paul.kienzle@nist.gov',
    url='http://github.com/reductus/reductus',
    description='Data reduction for neutron scattering',
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: Public Domain',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Physics',
    ],
    zip_safe=False,
    packages=packages,
    include_package_data=True,
    entry_points = {
        'console_scripts': ['reductus=reflweb.run:main'],
    },
    install_requires=[
        'scipy', 'numpy', 'h5py', 'uncertainties', 'docutils', 'wheel', 'pytz', 'msgpack-python', 'flask'],
    extras_require={
        'masked_curve_fit': ['numdifftools'],
        },
    )

# End of file
